import numpy as np
import pandas as pd
import time
import os
import pickle

from sklearn.model_selection import StratifiedKFold, ParameterGrid, train_test_split
# from sklearn.preprocessing import StandardScaler

from .. import constants as cnt
from .. import metrics as metr
from ..external import LeafModel

try:
    import mgzip  # Custom
    from memory_profiler import memory_usage
    open_file = mgzip.open
    check_memory = lambda: memory_usage()[0]
except ImportError:
    open_file = open
    check_memory = lambda: np.nan


def to_str_from_dict_list(d, stratify):
    if isinstance(stratify, str):
        return str(d.get(stratify, ""))
    elif isinstance(stratify, list):
        return ";".join([str(d.get(e, "")) for e in stratify])
    return None


def stratify_by_time_cens(y):
    qs = np.quantile(y[cnt.TIME_NAME], np.linspace(0.2, 0.8, 4))
    time_discr = np.searchsorted(qs, y[cnt.TIME_NAME])
    discr = np.char.add(time_discr.astype(str), y[cnt.CENS_NAME].astype(str))
    return discr


def prepare_sample(X, y, train_index, test_index):
    """ Constructing a set of bins on target variables and clipping """
    X_train, X_test = X.iloc[train_index, :].copy(), X.iloc[test_index, :].copy()
    y_train, y_test = y[train_index], y[test_index]

    # too_late = np.quantile(y_train[cnt.TIME_NAME], 0.975)
    # ind_tr = np.where(y_train[cnt.TIME_NAME] > too_late)
    # y_train[cnt.CENS_NAME][ind_tr] = False
    # y_train[cnt.TIME_NAME][ind_tr] = too_late

    # ind_tst = np.where(y_test[cnt.TIME_NAME] > too_late)
    # y_test[cnt.CENS_NAME][ind_tst] = False
    # y_test[cnt.TIME_NAME][ind_tst] = too_late

    bins = cnt.get_bins(time=y_train[cnt.TIME_NAME], cens=y_train[cnt.CENS_NAME])
    # y_train[cnt.TIME_NAME] = np.clip(y_train[cnt.TIME_NAME], bins.min() - 1, bins.max() + 1)

    y_test[cnt.TIME_NAME] = np.clip(y_test[cnt.TIME_NAME], bins.min(), bins.max())
    return X_train, y_train, X_test, y_test, bins


def generate_sample(X, y, folds, mode="CV"):
    """
    Generate cross-validate samples with StratifiedKFold.

    Parameters
    ----------
    X : Pandas dataframe
        Contain input features of events.
    y : structured array
        Contain censuring flag and time of events.
    folds : int
        Quantity of cross-validate folds.
    mode : str
        Validation scenario.

    Yields
    ------
    X_train : Pandas dataframe
        Contain input features of train sample.
    y_train : array-like
        Contain censuring flag and time of train sample.
    X_test : Pandas dataframe
        Contain input features of test sample.
    y_test : array-like
        Contain censuring flag and time of test sample.
    bins : array-like
        Points of timeline.

    """

    discr = stratify_by_time_cens(y)
    if mode != "HOLD-OUT":
        skf = StratifiedKFold(n_splits=folds)

    if mode == "TIME-CV":
        train_index = np.array([], dtype=int)
        for train_index_, test_index_ in skf.split(X, discr):  # y[cnt.CENS_NAME]):
            if train_index.shape[0] > 0:
                X_train, y_train, X_test, y_test, bins = prepare_sample(X, y, train_index, test_index_)
                yield X_train, y_train, X_test, y_test, bins
            train_index = np.hstack([train_index, test_index_])
    elif mode == "CV+HOLD-OUT":
        X, y, X_HO, y_HO, bins_HO = next(generate_sample(X, y, folds=1, mode="HOLD-OUT"))
        for X_train, y_train, X_test, y_test, bins in generate_sample(X, y, folds=folds, mode="CV"):
            yield X_train, y_train, X_test, y_test, bins
        yield X, y, X_HO, y_HO, bins_HO
    elif mode == "HOLD-OUT":
        for i_fold in range(folds):
            X_TR, X_HO = train_test_split(X, stratify=discr,  # y[cnt.CENS_NAME],
                                          test_size=0.33, random_state=42 + i_fold)
            X_tr, y_tr, X_HO, y_HO, bins_HO = prepare_sample(X, y, X_TR.index, X_HO.index)
            yield X_tr, y_tr, X_HO, y_HO, bins_HO
    elif mode == "CV":
        for train_index, test_index in skf.split(X, discr):  # y[cnt.CENS_NAME]):
            X_train, y_train, X_test, y_test, bins = prepare_sample(X, y, train_index, test_index)
            yield X_train, y_train, X_test, y_test, bins
    pass


def count_metric(y_train, y_test, pred_time, pred_sf, pred_hf, bins, metrics_names):
    return np.array([metr.METRIC_DICT[metr_name](y_train, y_test, pred_time, pred_sf, pred_hf, bins)
                     for metr_name in metrics_names])


def get_name_file(method, params, mode, fold):
    """
    Creating a name to cache the model without considering variables independent for reproducibility.
    """
    filter_params = ["categ", "ens_metric_name", "aggreg_func", "n_jobs"]
    name_lst = [method.__name__]
    name_lst += [v for k, v in params.items() if not (k in filter_params)]
    name_lst += [mode, fold]
    return "_".join(map(str, name_lst))


def get_fit_eval_func(method, X, y, folds, metrics_names=['CI'], mode="CV", dir_path=None):
    """
    Return function, which on sample X, y apply cross-validation and calculate 
    metrics for each fold.

    Parameters
    ----------
    method : class
        Must have methods for fitting, predicting time, hazard and survival func
            
    X : Pandas dataframe
        Contain input features of events.
    y : structured array
        Contain censuring flag and time of events.
    folds : int
        Quantity of cross-validate folds.
    metrics_names : TYPE, optional
        DESCRIPTION. The default is ['CI'].
    mode : str
        Validation scenario.
    dir_path : str
        Path to cache directory (for loading pretrained models).

    Returns
    -------
    functions
        Recieve hyperparameters and return list of metrics arrays.
        Allow to use in ParameterGrid.

    """
    def f(**kwargs):
        metr_lst = []
        exec_times = []
        exec_mem = []
        fold = 0
        for X_train, y_train, X_test, y_test, bins in generate_sample(X, y, folds, mode):
            # print("X_train.shape:", X_train.shape)
            s_time = time.time()
            s_mem = [check_memory()]
            est = method(**kwargs)
            if method.__name__.find('CRAID') != -1:  # TODO replace to isinstance
                if dir_path is None:
                    est.fit(X_train, y_train)
                else:
                    name = os.path.join(dir_path, get_name_file(method, kwargs, mode, fold) + '.pkl')
                    if not os.path.exists(name):
                        print("Fitted from scratch")
                        est.fit(X_train, y_train)
                        with open_file(name, 'wb') as out:  # Custom
                            pickle.dump(est, out, pickle.HIGHEST_PROTOCOL)
                    with open_file(name, 'rb') as inp:
                        est = pickle.load(inp)

                est.aggreg_func = kwargs.get("aggreg_func", "mean")
                est.tolerance_find_best(kwargs.get("ens_metric_name", "IBS_REMAIN"))
                s_mem.append(check_memory())
                pred_sf = est.predict_at_times(X_test, bins=bins, mode="surv")
                pred_sf[:, -1] = 0
                pred_sf[:, 0] = 1
                pred_time = est.predict(X_test, target=cnt.TIME_NAME)
                pred_hf = est.predict_at_times(X_test, bins=bins, mode="hazard")
            elif isinstance(est, LeafModel):
                X_train.loc[:, cnt.TIME_NAME] = y_train[cnt.TIME_NAME]
                X_train.loc[:, cnt.CENS_NAME] = y_train[cnt.CENS_NAME]
                est.fit(X_train)
                pred_sf = est.predict_survival_at_times(X_test, bins=bins)
                pred_time = est.predict_feature(X_test, feature_name=cnt.TIME_NAME)
                pred_hf = est.predict_hazard_at_times(X_test, bins=bins)
            else:  # Methods from scikit-survival
                s = pd.isna(X_train).sum(axis=0) != X_train.shape[0]
                valid_feat = s[s].index
                med_val = 0
                # med_val = X_train[valid_feat].median()
                # med_val = X_train[valid_feat].mean()
                X_train = X_train[valid_feat].fillna(med_val).replace(np.nan, med_val).replace(np.inf, med_val)
                X_test = X_test[valid_feat].fillna(med_val).replace(np.nan, med_val).replace(np.inf, med_val)

                # scaler = StandardScaler()
                # X_train = scaler.fit_transform(X_train)
                # X_test = scaler.transform(X_test)

                est = est.fit(X_train, y_train)
                s_mem.append(check_memory())
                survs = est.predict_survival_function(X_test)
                hazards = est.predict_cumulative_hazard_function(X_test)
                pred_sf = np.array(list(map(lambda x: x(bins), survs)))
                pred_hf = np.array(list(map(lambda x: x(bins), hazards)))
                pred_time = -1 * est.predict(X_test)
                # pred_time = np.trapz(pred_sf, bins)
                # Integral version from: https://lifelines.readthedocs.io/en/latest/fitters/regression/CoxPHFitter.html

            exec_mem.append(max(s_mem) - min(s_mem))
            exec_times.append(time.time() - s_time)
            metr_lst.append(count_metric(y_train, y_test, pred_time,
                                         pred_sf, pred_hf, bins, metrics_names))
            fold += 1
            del est
        return np.vstack(metr_lst), np.array(exec_times), np.array(exec_mem)
    return f


def bins_scheme(val, scheme=""):
    if scheme == "rank":
        u = np.unique(val)
        ind = np.digitize(val, u)
        return (ind * val.max() / ind.max()).astype("int")
    if scheme == "quantile":
        u = np.unique(np.quantile(val, np.linspace(0, 1, 100)))
        ind = np.digitize(val, u)
        return (ind * val.max() / ind.max()).astype("int")
    if scheme == "log+scale":
        a = np.log(val+1)
        v = (val.max() - val.min())*(a - a.min())/(a.max() - a.min()) + val.min()
        return v.astype("int")
    return val


class Experiments(object):
    """
    Class receives methods, metrics and grids,
          produces cross-validation experiments,
          stores table of results : name, params, time, metrics (by sample and mean)

    Attributes
    ----------
    methods : list
        Must have predicting methods according to metrics:
            IBS - survival func
            IAUC - cumulative hazard func
            CI - occurred time
            CI_CENS - occurred time
    methods_grid : list
        Each grid is dictionary: key - param name, values - list
    metrics : list
        Each metric is string, which must be in METRIC_DICT
    is_table : boolean
        Flag of calculation ending
    folds : int
        Quantity of cross-validate folds.
    except_stop : bool
        Mode of ending because of exception.
        True - stop experiments with current method
        False - continue experiments
    dataset_name : str
        Unique name of current dataset (used for saving)

    Methods
    -------
    
    add_method : append method and its grid
    set_metrics : check and set list of metric name 
    run : start experiments with data X, y
    get_agg_results : choose for each method aggregated params by metric and aggreg
    save : export table as xlsx
    
    """
    def __init__(self, folds=5, except_stop=False, dataset_name="NONE_NAME", mode="CV", bins_sch=""):
        self.methods = []
        self.methods_grid = []
        self.metrics = ["CI"]
        self.metric_best_p = "IBS"
        self.way_best_p = "min"

        self.is_table = False
        self.folds = folds
        self.except_stop = except_stop
        self.dataset_name = dataset_name
        self.dir_path = None

        self.result_table = None
        self.sample_table = None
        self.mode = mode
        self.bins_sch = bins_sch

    def add_metric_best(self, metric):
        if metric in self.metrics:
            self.metric_best_p = metric
            self.way_best_p = "min" if metric in metr.DESCEND_METRICS else "max"

    def add_method(self, method, grid):
        self.methods.append(method)
        self.methods_grid.append(grid)
        
    def set_metrics(self, lst_metric):
        self.metrics = []
        for metr_name in lst_metric:
            if metr_name in metr.METRIC_DICT:
                self.metrics.append(metr_name)
            else:
                print(f"METRIC {metr_name} IS NOT DEFINED")
    
    def run(self, X, y, dir_path=None, verbose=0):
        self.dir_path = dir_path
        y["time"] = bins_scheme(y["time"], scheme=self.bins_sch)
        self.result_table = pd.DataFrame([], columns=["METHOD", "PARAMS", "TIME"] + self.metrics)

        for method, grid in zip(self.methods, self.methods_grid):
            fit_eval_func = get_fit_eval_func(method, X, y, self.folds, self.metrics, self.mode, self.dir_path)
            print(method, grid)

            grid_params = ParameterGrid(grid)
            p_size = len(grid_params)
            for i_p, p in enumerate(grid_params):
                try:
                    eval_metr, exec_times, exec_mem = fit_eval_func(**p)
                    curr_dict = {"METHOD": method.__name__, "CRIT": p.get("criterion", ""), "PARAMS": str(p),
                                 "TIMES": exec_times, "TIME": np.sum(exec_times),
                                 "MEMS": exec_mem, "MEM": np.sum(exec_mem)}
                    eval_metr = {m: eval_metr[:, i] for i, m in enumerate(self.metrics)}
                    curr_dict.update(eval_metr)
                    self.result_table = pd.concat([self.result_table, pd.DataFrame([curr_dict])], ignore_index=True)
                    if verbose > 0:
                        print(f"Iteration: {i_p + 1}/{p_size}")
                        print(f"EXECUTION TIME OF {method.__name__}: {exec_times.round(3)}, MEM {exec_mem.round(3)}",
                              {k: [round(np.mean(v), 3), round(v[-1], 3)] for k, v in eval_metr.items()})
                except KeyboardInterrupt:
                    print("Handled KeyboardInterrupt")
                    break
                except Exception as e:
                    print("Method: %s, Param: %s finished with except '%s'" % (method.__name__, str(p), e))
                    if self.except_stop:
                        break
                    curr_dict = {"METHOD": method.__name__, "CRIT": p.get("criterion", ""),
                                 "PARAMS": str(p), "TIME": -1}
                    curr_dict.update({m: np.array([np.nan, np.nan]) for i, m in enumerate(self.metrics)})
                    self.result_table = pd.concat([self.result_table, pd.DataFrame([curr_dict])], ignore_index=True)
        if self.mode in ["TIME-CV", "CV+HOLD-OUT"]:
            for m in self.metrics:
                self.result_table[f"{m}_pred_mean"] = self.result_table[m].apply(lambda x: np.mean(x[:-1]))
            for m in self.metrics:
                self.result_table[f"{m}_last"] = self.result_table[m].apply(lambda x: x[-1])

        for m in self.metrics:
            self.result_table[f"{m}_mean"] = self.result_table[m].apply(np.mean)

        self.is_table = True

    def run_effective(self, X, y, dir_path=None, verbose=0,
                      stratify_best=["criterion", "balance", "leaf_model", "l_reg"]):
        if not (self.mode in ["CV+SAMPLE"]):
            self.run(X, y, dir_path=dir_path, verbose=verbose)
            return None

        y["time"] = bins_scheme(y["time"], scheme=self.bins_sch)
        self.bins_sch = ""

        folds = 20 if self.mode == "CV+SAMPLE" else 1

        discr = stratify_by_time_cens(y)
        X_TR, X_HO = train_test_split(X, stratify=discr,  # y[cnt.CENS_NAME],
                                      test_size=0.33, random_state=42)
        X_tr, y_tr, X_HO, y_HO, bins_HO = prepare_sample(X, y, X_TR.index, X_HO.index)
        old_mode = self.mode
        self.mode = "CV"
        self.run(X_tr, y_tr, dir_path=dir_path, verbose=verbose)
        self.sample_table = self.eval_on_sample_by_best_params(X, y, folds=folds, stratify=stratify_best)
        self.mode = old_mode

    def eval_on_sample_by_best_params(self, X, y, folds=20, stratify="criterion"):
        best_table = self.get_best_by_mode(stratify=stratify)
        d = best_table.loc[:, ["METHOD", "PARAMS"]].to_dict(orient="tight")
        map_method_by_name = {m.__name__: m for m in self.methods}

        ho_exp = Experiments(folds=folds, dataset_name=self.dataset_name, mode="HOLD-OUT")
        ho_exp.set_metrics(self.metrics)
        for method_params in d["data"]:
            method = map_method_by_name[method_params[0]]
            params = eval(method_params[1])
            params = {k: [v] for k, v in params.items()}
            ho_exp.add_method(method, params)
        ho_exp.run(X, y, dir_path=self.dir_path, verbose=1)
        res_table = ho_exp.result_table.copy()
        for m in self.metrics:
            res_table[f"{m}_CV"] = best_table[m]
            res_table[f"{m}_CV_mean"] = best_table[f"{m}_mean"]
        return res_table

    @staticmethod
    def get_agg_results(result_table, by_metric, choose="median", stratify="criterion"):
        if not (by_metric in result_table.columns):
            return None
        df = result_table.copy()
        stratify_name = f"Stratify({stratify})"
        df[stratify_name] = df["PARAMS"].apply(lambda x: to_str_from_dict_list(eval(x), stratify))
        df["METHOD_FULL"] = df.apply(lambda x: x["METHOD"].replace("CRAID", f"Tree({x[stratify_name]})"), axis=1)

        best_table = pd.DataFrame([], columns=df.columns)
        for method in df['METHOD_FULL'].unique():
            sub_table = df[df["METHOD_FULL"] == method]
            if sub_table.shape[0] == 0:
                continue
            if choose == "max":
                best_ind = sub_table[by_metric].idxmax()
                if np.isnan(best_ind):
                    continue
                best_row = sub_table.loc[best_ind]
            elif choose == "min":
                best_ind = sub_table[by_metric].idxmin()
                if np.isnan(best_ind):
                    continue
                best_row = sub_table.loc[best_ind]
            else:
                best_row = sub_table.sort_values(by=by_metric).iloc[sub_table.shape[0] // 2]

            # best_table = best_table.append(dict(best_row), ignore_index=True)
            best_table = pd.concat([best_table, pd.DataFrame([dict(best_row)])], ignore_index=True)
        return best_table

    def get_cv_result(self, stratify="criterion"):
        df_cv_best = self.get_agg_results(self.result_table, self.metric_best_p + "_mean",
                                          choose=self.way_best_p, stratify=stratify)
        return df_cv_best

    def get_time_cv_result(self, stratify="criterion"):
        df_time_cv_best = self.get_agg_results(self.result_table, self.metric_best_p + "_pred_mean",
                                               choose=self.way_best_p, stratify=stratify)
        return df_time_cv_best

    def get_hold_out_result(self, stratify="criterion"):
        df_hold_out_best = self.get_agg_results(self.result_table, self.metric_best_p + "_pred_mean",
                                                choose=self.way_best_p, stratify=stratify)
        rename_d = {m + "_pred_mean": m + "_CV_mean" for m in self.metrics}
        rename_d.update({m + "_last": m + "_HO" for m in self.metrics})
        return df_hold_out_best.rename(rename_d, axis=1)

    def get_result(self):
        return self.result_table

    def get_sample_result(self):
        return self.sample_table

    def get_best_by_mode(self, stratify="criterion"):
        if self.mode == "CV":
            return self.get_cv_result(stratify=stratify)
        elif self.mode == "TIME-CV":
            return self.get_time_cv_result(stratify=stratify)
        elif self.mode in ["CV+HOLD-OUT"]:
            return self.get_hold_out_result(stratify=stratify)
        elif self.mode in ["CV+SAMPLE"]:
            return self.get_sample_result()
        return None

    def save(self, dir_path):
        self.result_table.to_excel(f"{dir_path + self.dataset_name}_FULL_TABLE.xlsx", index=False)
            
    # def plot_results(self, dir_path):
    #     df = self.result_table.copy()
    #     df['METHOD'] = df.apply(lambda x: x["METHOD"].replace("CRAID", "Tree(%s)" %(x['CRIT'])),
    #                                  axis = 1)
    #     for m in self.metrics:
    #         fig, axs = plt.subplots(1, figsize=(6, 9))
    #         plt.title("%s %s" % (self.dataset_name, m))
    #         plt.boxplot(df[m], labels = df['METHOD'])
    #         plt.xticks(rotation=90)
    #         plt.savefig(dir_path + self.dataset_name + "%s_boxplot.png" %(m))
    #         plt.close(fig)



from .. import constants as cnt
from .leaf_model import LeafModel
from lifelines import WeibullAFTFitter, LogNormalAFTFitter, LogLogisticAFTFitter
from lifelines import CoxPHFitter
import numpy as np


PARAM_GRID = {
    "penalizer": [0, 0.01, 0.1, 0.5, 1.0],
    "l1_ratio": [100, 10, 1, 0.1, 0.01, 0.001]
}

AFT_param_grid = PARAM_GRID.copy()
CoxPH_param_grid = PARAM_GRID.copy()


class ParametricLifelinesBase(LeafModel):
    base_model = None

    def __init__(self, except_stop=False, **kwargs):
        self.model = None
        self.except_stop = except_stop
        super().__init__(**kwargs)

    def prepare_data(self, X):
        if X is None:
            if self.model is None:
                raise Exception("Model does not exist")
            return self.model._central_values

        if self.model is None:  # Prepare train sample
            self.features = set(self.features) & set((X.notna().sum() != 0).index)  # Delete full nan
            self.features = set(self.features) & set((X.std() == 0).index)  # Delete columns with one value
            self.features = list(self.features | {cnt.TIME_NAME, cnt.CENS_NAME})
            if X[cnt.CENS_NAME].sum() == 0:
                X[cnt.CENS_NAME] = 1  # parametric models do not work with only censored observations
        X = X[list(set(self.features) & set(X.columns))].copy()
        X = X.fillna(0).replace(np.nan, 0)
        return X

    def fit(self, X_node):
        super().fit(X_node)
        X_node = self.prepare_data(X_node)
        self.model = self.base_model(**self.kwargs)
        try:
            self.model.fit(X_node, cnt.TIME_NAME, event_col=cnt.CENS_NAME)
        except Exception as e:
            if self.except_stop:
                raise Exception(f"Leaf model fitting raise exception: {str(e)}")
            self.features = [cnt.TIME_NAME, cnt.CENS_NAME]
            self.model.fit(X_node[self.features], cnt.TIME_NAME, event_col=cnt.CENS_NAME)

    def predict_survival_at_times(self, X=None, bins=None):
        X = self.prepare_data(X)
        return self.model.predict_survival_function(X, times=bins).to_numpy().T

    def predict_hazard_at_times(self, X=None, bins=None):
        X = self.prepare_data(X)
        chf = self.model.predict_cumulative_hazard(X, times=bins).to_numpy().T
        return chf

    def predict_feature(self, X=None, feature_name=None):
        if feature_name == cnt.TIME_NAME:
            ts = np.linspace(self.model.durations.min(), self.model.durations.max(), 100)
            sf = self.predict_survival_at_times(X=X, bins=ts)
            return np.trapz(sf, ts, axis=1)
            # return self.model.predict_expectation(X).to_numpy()
        X = self.prepare_data(X)
        return super().predict_feature(X, feature_name)


class WeibullAFT(ParametricLifelinesBase):
    base_model = WeibullAFTFitter


class LogNormalAFT(ParametricLifelinesBase):
    base_model = LogNormalAFTFitter


class LogLogisticAFT(ParametricLifelinesBase):
    base_model = LogLogisticAFTFitter


class CoxPH(ParametricLifelinesBase):
    base_model = CoxPHFitter


LEAF_AFT_DICT = {
    "WeibullAFT": WeibullAFT,
    "LogNormalAFT": LogNormalAFT,
    "LogLogisticAFT": LogLogisticAFT,
    "CoxPH": CoxPH
}



import numpy as np
import pandas as pd
import re
import pickle 
import random
from os.path import dirname, join
from ..constants import TIME_NAME, CENS_NAME, get_y
from sklearn import preprocessing

random.seed(10)

sign_gbsg = ['htreat', 'age', 'menostat', 'tumsize', 'tumgrad', 'posnodal', 'prm', 'esm']
categ_gbsg = ['htreat', 'menostat', 'tumgrad']

sign_pbc = ['trt', 'age', 'sex', 'ascites', 'hepato', 'spiders', 'edema', 'bili', 'chol',
            'albumin', 'copper', 'alk', 'ast', 'trig', 'platelet', 'protime', 'stage']
categ_pbc = ['trt', 'sex', 'ascites', 'hepato', 'spiders']


def save_pickle(obj, path):
    file_pi = open(path, 'wb') 
    pickle.dump(obj, file_pi, pickle.HIGHEST_PROTOCOL)


def load_pickle(path):
    return pickle.load(open(path, 'rb'))


def load_scheme_dataset(name):
    """
    Base loader function for internal datasets

    Parameters
    ----------
    name: str

    Returns
    -------
    X: pd.DataFrame
        Feature space
    y: structured np.ndarray
        Target variables
    sign: list
        All variables
    categ: list
        Categorical variables
    sch_nan: list
        Scheme variables (supported multiple prediction)

    """
    dir_env = join(dirname(__file__), "data")
    df = pd.read_csv(join(dir_env, f'{name}.csv'))
    df = df.rename({"time": TIME_NAME, "event": CENS_NAME}, axis=1)

    sign_c = [c for c in df.columns if c.startswith("num_") or c.startswith("fac_")]
    categ_c = [c for c in df.columns if c.startswith("fac_")]

    y = get_y(cens=df[CENS_NAME], time=df[TIME_NAME].astype("int"))
    X = df.loc[:, sign_c]

    for c in categ_c:
        le = preprocessing.LabelEncoder()
        X[c] = le.fit_transform(X[c])

    if y[TIME_NAME].min() == 0:
        y[TIME_NAME] += 1
    return X, y, sign_c, categ_c, []


def load_support2_dataset():
    """
    Full description: https://archive.ics.uci.edu/dataset/880/support2
    """
    return load_scheme_dataset('support2')

def load_rott2_dataset():
    """
    Full description: https://rdrr.io/cran/AF/man/rott2.html
    """
    return load_scheme_dataset('rott2')

def load_Framingham_dataset():
    """
    Full description: https://search.r-project.org/CRAN/refmans/riskCommunicator/html/framingham.html
    """
    return load_scheme_dataset('Framingham')

def load_flchain_dataset():
    """
    Full description: https://pmagunia.com/dataset/r-dataset-package-survival-flchain
    """
    return load_scheme_dataset('flchain')

def load_smarto_dataset():
    """
    Full description: https://pubmed.ncbi.nlm.nih.gov/36806141/
    """
    return load_scheme_dataset('smarto')

def load_actg_dataset():
    """
    Full description: https://rdrr.io/cran/mlr3proba/man/actg.html
    """
    return load_scheme_dataset("actg")


def load_gbsg_dataset():
    """
    Full description: https://paperswithcode.com/dataset/gbsg2
    """
    dir_env = join(dirname(__file__), "data")
    df = pd.read_csv(join(dir_env, 'GBSG.csv'))
    df = df.rename({"rfst": TIME_NAME, "cens": CENS_NAME}, axis=1)
    
    y = get_y(cens=df[CENS_NAME], time=df[TIME_NAME])
    X = df.loc[:, sign_gbsg]

    if y[TIME_NAME].min() == 0:
        y[TIME_NAME] += 1
    return X, y, sign_gbsg, categ_gbsg, []


def load_pbc_dataset():
    """
    Full description: https://stat.ethz.ch/R-manual/R-patched/library/survival/html/pbc.html
    """
    dir_env = join(dirname(__file__), "data")
    df = pd.read_csv(join(dir_env, 'pbc.csv'))
    df = df.rename({"time": TIME_NAME, "status": CENS_NAME, "alk.phos": "alk"}, axis=1)
    df[CENS_NAME] = np.array(df[CENS_NAME] > 1, dtype=int)
    df['sex'] = df['sex'].map({'f': 1, 'm': 0})
    
    y = get_y(cens=df[CENS_NAME], time=df[TIME_NAME])
    X = df.loc[:, sign_pbc]

    if y[TIME_NAME].min() == 0:
        y[TIME_NAME] += 1
    return X, y, sign_pbc, categ_pbc, []


def load_wuhan_dataset(invert_death=False):
    """
    Full description: https://www.nature.com/articles/s42256-020-0180-7
    """
    dir_env = join(dirname(__file__), "data")
    df = pd.read_excel(join(dir_env, 'covid_train.xlsx'))
    df['PATIENT_ID'] = df['PATIENT_ID'].fillna(method='ffill')
    columns_no_agg = ['RE_DATE', 'age', 'gender', 'Admission time', 'Discharge time', 'outcome']
    #  columns_agg = list(set(df.columns) - set(columns_no_agg))
    df_agg = df.groupby('PATIENT_ID').agg(list)
    for c in df_agg.columns:
        if c in columns_no_agg:
            df_agg[c] = df_agg[c].apply(lambda x: x[0])
        else:
            df_agg['mean_' + c] = df_agg[c].apply(np.nanmean)
            df_agg['min_' + c] = df_agg[c].apply(np.nanmin)
            df_agg['max_' + c] = df_agg[c].apply(np.nanmax)
            df_agg = df_agg.drop(c, axis=1)
    df_agg['time'] = df_agg.loc[:, ['Admission time', 'Discharge time']].apply(lambda x: (x['Discharge time'] - x['Admission time']).days, axis=1)
    df_f = df_agg[df_agg['time'] == df_agg['time']]
    df_f = df_f.drop(['RE_DATE', 'Admission time', 'Discharge time'], axis=1)
    df_f = df_f.rename({'outcome': CENS_NAME, 'time': TIME_NAME}, axis=1)
    df_f = df_f.rename({c: re.sub('[^A-Za-z0-9_]', '_', c) for c in df_f.columns}, axis=1)
    df_f = df_f.reset_index(drop=True)
    
    categ_covid = []
    sign_covid = sorted(list(set(df_f) - {CENS_NAME, TIME_NAME}))
    sign_covid = list(set(sign_covid) - {"max_2019_nCoV_nucleic_acid_detection",
                                         "mean_2019_nCoV_nucleic_acid_detection",
                                         "min_2019_nCoV_nucleic_acid_detection"})
    
    if invert_death:
        df_f[CENS_NAME] = 1 - df_f[CENS_NAME]
    y = get_y(cens=df_f[CENS_NAME], time=df_f[TIME_NAME])
    X = df_f.loc[:, sorted(sign_covid)]

    if y[TIME_NAME].min() == 0:
        y[TIME_NAME] += 1
    return X, y, sign_covid, categ_covid, []



import pandas as pd
from os.path import dirname, join

TIME_NAME = "time"
CENS_NAME = "cens"


def get_y(cens, time):
    cens, time = np.array(cens), np.array(time) 
    y = np.empty(dtype=[(CENS_NAME, bool), 
                        (TIME_NAME, np.float64)], 
                 shape=cens.shape[0])
    y[CENS_NAME] = cens
    y[TIME_NAME] = time
    return y


def str_to_categ(df_col):
    uniq = df_col.unique()
    return df_col.map(dict(zip(uniq, range(len(uniq)))))


def load_backblaze_2016_2018(threshold=0.99):
    df = pd.read_csv('./backblaze_drop_truncated_2016_2018.csv')
    df['time'] = pd.to_timedelta(df['time']).dt.days
    df = df.rename({"time": TIME_NAME, "event": CENS_NAME}, axis=1)
    nan_percentage = df.isna().mean()
    columns_with_high_nan = nan_percentage[nan_percentage >= threshold].index.tolist()
    df = df.drop(columns=['serial_number', 'date', 'time_row'] + columns_with_high_nan)
    categ = ['model']
    df['model'] = str_to_categ(df['model'])
    sign = sorted(list(set(df.columns) - {CENS_NAME, TIME_NAME}))
    y = get_y(df[CENS_NAME], df[TIME_NAME] + 1)
    X = df.loc[:, sign]
    return X, y, sign, categ


import numpy as np
import pandas as pd
from re import sub
from os.path import dirname, join
from ..constants import TIME_NAME, CENS_NAME, get_y

try:
    from dateutil import parser
except ImportError:
    raise ImportError('Please, install python-dateutil: pip install python-dateutil')


schemes_list = [
    'light_1', 'light_2', 'light_3',
    'middle_1', 'middle_2', 'middle_3', 'middle_4', 'middle_5', 'middle_6', 'middle_7', 'middle_8',
    'hard_1', 'hard_2', 'hard_3', 'hard_4', 'hard_5',
    'cyto_1', 'cyto_2', 'cyto_3', 'cyto_4', 'cyto_5', 'cyto_6',
    'MOS_AM', 'MOS_AM_AZ', 'N_MOS_AM_AZ', 'MOS_AM_LVF', 'N_MOS_AM_LVF',
    'MOS_CFTR', 'MOS_CFT', 'MOS_CFP', 'MOS_CFP_LVF', 'N_MOS_CFP_LVF',
    'MOS_MOK', 'MOS_PIP', 'MOS_LVF_MOK', 'mos_light_1', 'mos_light_2',
    'mos_middle_1', 'mos_middle_2', 'mos_middle_3', 'mos_middle_4',
    'mos_hard_1', 'mos_hard_2', 'mos_hard_3', 'mos_cyto_1', 'mos_cyto_2',
    # 'cheme_therap_nmg_hor_inter', 'cheme_therap_nmg_hor',
    # 'cheme_therap_nmg_inter', 'dexamethasone', 'methylprednisolone',
    # 'dex_meth'
]

important = [TIME_NAME, CENS_NAME, 'SUM_M']

categ_cyto = [
    'has_odyshka_ili_zatrudnennoe_dyhanie', 'has_oschussh_zalozhen_v_grudnoi_kletke',
    'has_slabost_ili_lomota', 'Пол', 'IGG_DEF', 'IGM_DEF', 'PCR_N',
    'сахарный диабет', 'ожирение', 'ишемическая болезнь сердца',
    'has_kashel', 'kashel_type', 'NIT', 'resultat_KT'
] + schemes_list

def_categ = {
    'osmotr_tyazhest': {'без симптомов': 0, 'легкая': 3, 'средняя': 5, 'тяжелая': 7},
    'has_odyshka_ili_zatrudnennoe_dyhanie': {'отсутствует': 0, 'присутствует': 1},
    'has_oschussh_zalozhen_v_grudnoi_kletke': {'отсутствует': 0, 'присутствует': 1},
    'has_slabost_ili_lomota': {'отсутствует': 0, 'присутствует': 1},
    'Пол': {'Мужской': 0, 'Женский': 1, 'м': 0, 'ж': 1, 'Нет данных': np.nan, 0: 0, 1: 1},
    'PCR_N': {'НЕ ОБНАРУЖЕНО': 0, 'ОБНАРУЖЕНО': 1},
    'has_kashel': {'отсутствует': 0, 'присутствует': 1},
    'kashel_type': {'сухой': 0, 'с мокротой': 1},
    'Группа риска': {'Хроники': 0, 'Беременные': 1, 'Старше 65': 2}
}

# True categorical dictionary (takes into account filled values)
def_categ_full = {
    'osmotr_tyazhest': {'без симптомов': 0, 'легкая': 3, 'средняя': 5, 'тяжелая': 7},
    'has_odyshka_ili_zatrudnennoe_dyhanie': {'отсутствует': 0, 'присутствует': 1, 0: 0, 1: 1},
    'has_oschussh_zalozhen_v_grudnoi_kletke': {'отсутствует': 0, 'присутствует': 1, 0: 0, 1: 1},
    'has_slabost_ili_lomota': {'отсутствует': 0, 'присутствует': 1, 0: 0, 1: 1},
    'Пол': {'Мужской': 0, 'Женский': 1, 'м': 0, 'ж': 1, 'Нет данных': np.nan, 0: 0, 1: 1},
    'PCR_N': {'НЕ ОБНАРУЖЕНО': 0, 'ОБНАРУЖЕНО': 1, 0: 0, 1: 1},
    'has_kashel': {'отсутствует': 0, 'присутствует': 1, 0: 0, 1: 1},
    'kashel_type': {'сухой': 0, 'с мокротой': 1, 0: 0, 1: 1},
    'Группа риска': {'Хроники': 0, 'Беременные': 1, 'Старше 65': 2, 0: 0, 1: 1, 2: 2}
}

sign = [
    # 'osmotr_tyazhest',
    'days_f', 'days_s', 'spo2', 'chdd', 'has_odyshka_ili_zatrudnennoe_dyhanie', 'temperatura_tela_value',
    'has_oschussh_zalozhen_v_grudnoi_kletke', 'has_slabost_ili_lomota', 'Пол', 'возраст',
    'IGG_N', 'IGM_N', 'IGG_DEF', 'IGM_DEF', 'PCR_N',
    'сахарный диабет', 'ожирение', 'ишемическая болезнь сердца',
    'Определение белков острой фазы С-реактивный белок',
    'has_kashel', 'kashel_type',
    'D-димер', 'Исследование ферритина',
    'NIT', 'АСТ', 'АЛТ', 'URO', 'Определение креатинина',
    'GRA#', 'Определение калия общего',
    'по Вестергрену', 'по Панченкову',
    'MON#',
    'HGB', 'MPV', 'HCT', 'PLT', 'PDW',
    'LYM#',
    'GLU', 'WBC', 'RBC',
    'RDW',
    'Абсолютное количество эозинофилов',
    'Абсолютное количество базофилов',
    'Абсолютное количество нейтрофилов',
    'resultat_KT'
] + schemes_list

therap_antic = "Гепарин Далтепарин Надропарин Эноксапарин Парнапарин Бемипарин Фондапаринукс Ривароксабан Апиксабан"
prevent_antic = therap_antic
glucoco = "Преднизолон Дексаметазон Метилпреднизолон Гидрокортизон"

mapp_sch = {
    "light_1": "Гидроксихлорохин",
    "light_2": "Мефлохин",
    "light_3": "ИФН Умифеновир",
    "middle_1": "Фавипиравир Барицитиниб Тофацитиниб",
    "middle_2": "Гидроксихлорохин Азитромицин Барицитиниб Тофацитиниб",
    "middle_3": "Мефлохин Азитромицин Барицитиниб Тофацитиниб",
    "middle_4": "Лопинавир Ритонавир ИФН Барицитиниб Тофацитиниб",
    "middle_5": "Фавипиравир Олокизумаб",
    "middle_6": "Гидроксихлорохин Азитромицин Олокизумаб",
    "middle_7": "Мефлохин Азитромицин Олокизумаб",
    "middle_8": "Лопинавир Ритонавир ИФН Олокизумаб",
    "hard_1": "Фавипиравир Тоцилизумаб Сарилумаб",
    "hard_2": "Гидроксихлорохин Азитромицин Тоцилизумаб Сарилумаб",
    "hard_3": "Мефлохин Азитромицин Тоцилизумаб Сарилумаб",
    "hard_4": "Лопинавир Ритонавир ИФН Тоцилизумаб Сарилумаб",
    "hard_5": "Лопинавир Ритонавир Гидроксихлорохин Тоцилизумаб Сарилумаб",
    "cyto_1": "Метилпреднизолон Тоцилизумаб Сарилумаб",
    "cyto_2": "Дексаметазон Тоцилизумаб Сарилумаб",
    "cyto_3": "Метилпреднизолон Канакинумаб",
    "cyto_4": "Дексаметазон Канакинумаб",
    "cyto_5": "Метилпреднизолон Дексаметазон",
    "cyto_6": "Тоцилизумаб Сарилумаб Канакинумаб",
    "MOS_AM": "Амоксициллин",
    "MOS_AM_AZ": "Ампициллин Азитромицин",     # Антибиотик: Ампициллин
    "N_MOS_AM_AZ": "Азитромицин",
    "MOS_AM_LVF": "Ампициллин Левофлоксацин",  # Антибиотик: Ампициллин
    "N_MOS_AM_LVF": "Левофлоксацин",
    "MOS_CFTR": "Цефтриаксон",                 # Антибиотик Цефтриаксон
    "MOS_CFT": "Цефотаксим",                   # Антибиотик Цефотаксим
    "MOS_CFP": "Цефепим",                      # Антибиотик Цефепим
    "MOS_CFP_LVF": "Цефепим Левофлоксацин",    # Антибиотик Цефепим
    "N_MOS_CFP_LVF": "Левофлоксацин",
    "MOS_MOK": "Моксифлоксацин",
    "MOS_PIP": "Пиперациллин",               # Антибиотик Пиперациллин

    "MOS_LVF_MOK": "Левофлоксацин Моксифлоксацин",
    'mos_light_1': "Фавипиравир Риамиловир Гидроксихлорохин ",
    'mos_light_2': "Фавипиравир Риамиловир Гидроксихлорохин " + therap_antic,
    'mos_middle_1': "Фавипиравир Риамиловир Гидроксихлорохин " + prevent_antic,
    'mos_middle_2': "Фавипиравир Риамиловир Гидроксихлорохин " + prevent_antic,
    'mos_middle_3': "Фавипиравир Риамиловир Гидроксихлорохин " + prevent_antic,
    'mos_middle_4': "Фавипиравир Риамиловир Гидроксихлорохин " + prevent_antic,
    'mos_hard_1': "Фавипиравир " + prevent_antic + glucoco,
    'mos_hard_2': "Риамиловир " + prevent_antic + glucoco,
    'mos_hard_3': "Гидроксихлорохин " + prevent_antic + glucoco,
    'mos_cyto_1': "Фондапаринукс " + glucoco,
    'mos_cyto_2': "Фондапаринукс " + prevent_antic + glucoco
}
map_bool = {'0': 'Нет', '1': 'Да'}


def create_none_ft(fill_df, sign_f, fill_none=False):
    """
    Fill nans in input DataFrame for definite features.
    Also, there is renumbering of categorical features
    (in dictionary def_categ or have a constraint for unique values).

    Parameters
    ----------
    fill_df : DataFrame
    sign_f : list
    fill_none : bool

    Returns
    -------
    DataFrame
    """
    for i in sign_f:
        # fill_df['none_'+i] = fill_df[i].apply(lambda x: int(x != x))
        if i in set(def_categ.keys()):
            # categorical features
            fill_df[i] = fill_df[i].map(def_categ[i])
            print('Categ-contin var:', i)
            if fill_none:
                fill_df[i] = fill_df[i].fillna(fill_df[i].median())
                fill_df[i] = fill_df[i].astype('int')
        else:
            # continuous features
            right_thres = 4
            if fill_none:
                right_thres = 3
                fill_df[i] = fill_df[i].fillna(fill_df[i].median())
                fill_df[i] = fill_df[i].astype('float64')
            len_uniq = len(fill_df[i].unique())
            if len_uniq < 10:
                if (len_uniq <= 6) and (len_uniq >= right_thres):
                    print('Categorical var:', i, len_uniq)
                    dict_encoder = {v_u: i_u for i_u, v_u in enumerate(np.sort(fill_df[i].unique())) if v_u == v_u}
                    fill_df[i] = fill_df[i].map(dict_encoder)
    return fill_df


def dest_date(x, y, with_neg=False):
    """
    Calculate the destination between dates.

    Parameters
    ----------
    x : str
    y : str
    with_neg : bool

    Returns
    -------
    int
    """
    days = np.nan
    try:
        p_1 = parser.parse(x)
        p_2 = parser.parse(y)
        days = (p_1 - p_2).days
        if not with_neg:
            days = abs(days)
    except Exception as e:
        pass
    return days


def load_covid_dataset(dir_env=None):
    if dir_env is None:
        dir_env = join(dirname(__file__), "data", "COVID")
    cyto = pd.read_csv(join(dir_env, "cyto_with_schemes.csv"))
    cyto = create_none_ft(cyto, sign)
    cyto[TIME_NAME] = cyto.apply(lambda row: dest_date(row['Дата создания ЭС'], row['Дата закрытия ЭС']), axis=1)
    cyto[CENS_NAME] = cyto['Причина закрытия ЭС'].apply(lambda x: int(x[0] == 'У') if x == x else 0)
    cyto['SUM_M'] = 0
    sec = cyto.loc[:, sign+important]
    
    repl = {sc: sub("[^0-9A-Za-zА-Яа-я_]", "", sc) for sc in sec.columns}
    sec = sec.rename(columns=repl)
    sec['idx'] = sec.index 
    sec = sec[sec[TIME_NAME] == sec[TIME_NAME]]
    for sch in schemes_list:
        sec[sch] = sec[sch].fillna(0)
    new_sign = [repl[s] for s in sign]
    new_categ = [repl[s] for s in categ_cyto]
    new_important = [repl[s] for s in schemes_list]
    
    sec = sec.reset_index()
    y = get_y(cens=sec[CENS_NAME], time=sec[TIME_NAME])
    X = sec.loc[:, new_sign]
    return X, y, new_sign, new_categ, new_important

### ADD SCHEMES

# def add_schemes():
#     cyto = pd.read_csv(dir_env + 'cyto_train_16.csv')
#     schemes_df = pd.read_csv(dir_env + 'second_all_schemes.csv')
#     simi = pd.read_excel(dir_env + 'simi_uuid.xlsx')
#     cyto_global = cyto.merge(simi, how = 'left', on = 'uuid')
#     cyto_global['kis_uuid'] = cyto_global.apply(lambda x: x['kis_uuid'] if x['kis_uuid'] == x['kis_uuid'] else x['uuid'],axis = 1)
#     schemes_df = schemes_df.rename({'id':'kis_uuid'},axis = 1)
#     schemes_df = schemes_df.loc[:,['kis_uuid']+schemes_list]#.sum(axis = 1).value_counts()
#     cyto_global = cyto_global.merge(schemes_df, how = 'left', on = 'kis_uuid')
#     cyto_global.to_csv(dir_env + 'cyto_with_schemes.csv')



import pandas as pd
import numpy as np
from sklearn.metrics import roc_auc_score

from .. import metrics as metr
from .. import constants as cnt


class BaseEnsemble(object):
    """
    Base ensemble of survival decision tree.

    Attributes
    ----------
    size_sample : float
        Size of generated subsample
    n_estimators : int
        Number of base models
    ens_metric_name : str
        Metric defines quantitative of ensemble
    descend_metr : boolean
        Flag of descend for ens_metric_name
    bootstrap : boolean
        Flag for using bootstrap sampling (with return)
    tolerance : boolean
        Flag for fitting full ensemble and choosing best submodels
    aggreg : boolean
        Flag of aggregating base responses
    aggreg_func : str
        Function of aggregating (if aggreg)
    tree_kwargs : dict
        Parameters for building base models

    models : list
        Base models of ensemble (for example, CRAID)
    oob : list
        Out of bag sample for each model
    ens_metr : array-like
        Values of ens_metric_name for ensemble with i models
    list_pred_oob : list
        Predictions of ensemble models[:i+1]

    Methods
    -------
    update_params : attributes preparation
    fit : build ensemble with X, y data (abstract)
    add_model : updating ensemble with new model and oob
    select_model : remaining fixed models
    tolerance_find_best : iterative method of selecting best sub ensemble

    predict : return values of features, rules or schemes (look at CRAID)
    predict_at_times : return survival or hazard function

    score_oob : calculate metric "ens_metric_name" for ensemble
    """

    def __init__(self, size_sample=0.7,
                 n_estimators=10,
                 aggreg=True,
                 ens_metric_name="roc",
                 bootstrap=True,
                 tolerance=True,
                 aggreg_func="mean",
                 **tree_kwargs):
        self.size_sample = size_sample
        self.n_estimators = n_estimators
        self.aggreg = aggreg
        # self.ens_metric_name = ens_metric_name
        # self.descend_metr = self.ens_metric_name in metr.DESCEND_METRICS
        self.bootstrap = bootstrap
        self.tolerance = tolerance
        self.tree_kwargs = tree_kwargs
        self.aggreg_func = aggreg_func
        self.X_train = None
        self.y_train = None
        self.features = None

    def update_params(self):
        self.models = []
        self.oob = []
        self.ens_metr = np.zeros(self.n_estimators)
        self.list_pred_oob = []

        if isinstance(self.size_sample, float):
            self.size_sample = int(self.size_sample * self.X_train.shape[0])
        self.bins = cnt.get_bins(time=self.y_train[cnt.TIME_NAME],
                                 cens=self.y_train[cnt.CENS_NAME])

        cnt.set_seed(10)

    def fit(self):
        pass

    def add_model(self, model, x_oob):
        self.models.append(model)
        self.oob.append(x_oob)
        if self.ens_metric_name == "CI":
            tree_pred = pd.DataFrame(model.predict(x_oob, target=cnt.TIME_NAME), index=x_oob.index)
        elif self.ens_metric_name == "roc":
            tree_pred = pd.DataFrame(model.predict(x_oob, target=cnt.CENS_NAME), index=x_oob.index)
        else:
            tree_pred = pd.DataFrame(model.predict_at_times(x_oob, bins=self.bins, mode="surv"), index=x_oob.index)
            tree_pred = tree_pred.apply(lambda r: np.array(r), axis=1)
        self.list_pred_oob.append(tree_pred)

    def select_model(self, start, end):
        self.models = self.models[start:end]
        self.oob = self.oob[start:end]
        self.list_pred_oob = self.list_pred_oob[start:end]

    def tolerance_find_best(self):
        if self.descend_metr:
            max_index = np.argmin(self.ens_metr)
        else:
            max_index = np.argmax(self.ens_metr)
        self.select_model(0, max_index + 1)
        print(self.ens_metr)

    def get_aggreg(self, x):
        if self.aggreg_func == "median":
            return np.median(x, axis=0)
        elif self.aggreg_func == "mean":
            return np.mean(x, axis=0)
        return None

    def predict_at_times(self, x_test, bins, aggreg=True, mode="surv"):
        res = []
        for i in range(len(self.models)):
            res.append(self.models[i].predict_at_times(x_test, bins=bins,
                                                       mode=mode)[np.newaxis, :])
        res = np.vstack(res)
        if aggreg:
            res = self.get_aggreg(res)
            if mode == "surv":
                res[:, -1] = 0
                res[:, 0] = 1
        return res

    def predict(self, x_test, aggreg=True, **kwargs):
        res = []
        for i in range(len(self.models)):
            res.append(self.models[i].predict(x_test, **kwargs))
        res = np.array(res)
        if aggreg:
            res = self.get_aggreg(res)
        return res

    """ SCORES """

    def score_oob(self):
        if self.aggreg:
            score = self.aggregate_score_selfoob()
        else:
            score = self.separate_score_oob()
        return np.round(score, 4)

    def separate_score_oob(self):
        scores = np.array([])
        for i in range(len(self.models)):
            X_v = self.oob[i]
            if self.ens_metric_name == "CI":
                pred = self.models[i].predict(X_v, target=cnt.TIME_NAME)
                score = metr.concordance_index(X_v[cnt.TIME_NAME], pred)
            else:
                pred = self.models[i].predict_at_times(X_v, bins=self.bins, mode="surv")
                y_true = cnt.get_y(X_v[cnt.CENS_NAME], X_v[cnt.TIME_NAME])
                score = metr.ibs(self.y_train, y_true, pred, self.bins)
            scores = np.append(scores, score)
        return np.round(np.mean(scores), 4)

    def aggregate_score_selfoob(self, bins=None):
        is_ibs = self.ens_metric_name.upper().find("IBS") >= 0
        if self.ens_metric_name in ["CI"] or is_ibs:
            list_target_time = [oob_[cnt.TIME_NAME].to_frame() for oob_ in self.oob]
            target_time = pd.concat(list_target_time, axis=1).mean(axis=1)

        if self.ens_metric_name in ["roc"] or is_ibs:
            list_target_cens = [oob_[cnt.CENS_NAME].to_frame() for oob_ in self.oob]
            target_cens = pd.concat(list_target_cens, axis=1).mean(axis=1)

        if self.ens_metric_name in ["CI", "roc"]:
            pred = pd.concat(self.list_pred_oob, axis=1).mean(axis=1)
            if self.ens_metric_name == "CI":
                return metr.concordance_index(target_time, pred)
            return roc_auc_score(target_cens, pred)

        if is_ibs:
            pred = pd.concat(self.list_pred_oob, axis=1).apply(lambda r: r.mean(axis=0), axis=1)
            pred = np.array(pred.to_list())
            y_true = cnt.get_y(target_cens, target_time)
            return metr.ibs(self.y_train, y_true, pred, self.bins)
        return None


class FastBaseEnsemble(BaseEnsemble):
    def update_params(self):
        self.models = []
        self.oob_index = []

        if isinstance(self.size_sample, float):
            self.size_sample = int(self.size_sample * self.X_train.shape[0])
        self.bins = cnt.get_bins(time=self.y_train[cnt.TIME_NAME],
                                 cens=self.y_train[cnt.CENS_NAME])
        cnt.set_seed(10)

    def add_model(self, model, x_oob):
        self.models.append(model)
        self.oob_index.append(x_oob.index.values)

    def select_model(self, start, end):
        self.models = self.models[start:end]
        self.oob_index = self.oob_index[start:end]

    def tolerance_find_best(self, ens_metric_name="bic"):
        self.ens_metric_name = ens_metric_name

        self.is_ci = self.ens_metric_name.upper().find("CI") >= 0
        self.is_ibs = self.ens_metric_name.upper().find("IBS") >= 0
        self.is_auprc = self.ens_metric_name.upper().find("AUPRC") >= 0
        self.is_iauc = self.ens_metric_name.upper().find("IAUC") >= 0

        ens_metr_arr = np.zeros(self.n_estimators)

        self.prepare_for_tolerance()
        for i in range(self.n_estimators):
            self.tolerance_iter = i
            self.predict_by_i(i)
            ens_metr_arr[i] = self.score_oob()

        descend_metr = self.ens_metric_name in metr.DESCEND_METRICS
        if descend_metr:
            best_index = np.argmin(ens_metr_arr)
        else:
            best_index = np.argmax(ens_metr_arr)
        self.select_model(0, best_index + 1)
        # print(ens_metr_arr)
        # print(f"fitted: {len(self.models)} models.")

    def prepare_for_tolerance(self):
        if self.ens_metric_name in ["LOGLIKELIHOOD", "bic"] or self.is_ibs or self.is_iauc or self.is_auprc:
            dim = (self.X_train.shape[0], self.bins.shape[0])
        else:
            dim = (self.X_train.shape[0])
        self.oob_prediction = np.zeros(dim, dtype=float)

        if self.ens_metric_name in ["LOGLIKELIHOOD", "bic"]:
            self.oob_prediction_hf = np.zeros(dim, dtype=float)
        self.oob_count = np.zeros((self.X_train.shape[0]), dtype=int)

    def predict_by_i(self, ind_model):
        model = self.models[ind_model]
        oob_index = self.oob_index[ind_model]
        x_oob = self.X_train.iloc[oob_index, :]

        self.oob_count[oob_index] += 1
        if self.is_ci:
            self.oob_prediction[oob_index] += model.predict(x_oob, target=cnt.TIME_NAME)
        elif self.ens_metric_name == "roc":
            self.oob_prediction[oob_index] += model.predict(x_oob, target=cnt.CENS_NAME)
        elif self.ens_metric_name in ["LOGLIKELIHOOD", "bic"]:
            self.oob_prediction_hf += model.predict_at_times(self.X_train, bins=self.bins, mode="hazard")
            self.oob_prediction += model.predict_at_times(self.X_train, bins=self.bins, mode="surv")
        elif self.is_iauc:
            self.oob_prediction[oob_index] += model.predict_at_times(x_oob, bins=self.bins, mode="hazard")
        elif self.is_ibs or self.is_auprc:
            self.oob_prediction[oob_index] += model.predict_at_times(x_oob, bins=self.bins, mode="surv")

    def aggregate_score_selfoob(self):
        index_join_oob = np.where(self.oob_count != 0)

        pred_time = None
        pred_sf, pred_hf = None, None
        if self.is_ibs or self.is_auprc:
            pred_sf = self.oob_prediction[index_join_oob] / self.oob_count[index_join_oob][:, None]
        elif self.ens_metric_name in ["LOGLIKELIHOOD", "bic"]:
            pred_hf = self.oob_prediction_hf[index_join_oob] / (self.tolerance_iter + 1)
            pred_sf = self.oob_prediction[index_join_oob] / (self.tolerance_iter + 1)
        elif self.is_iauc:
            pred_hf = self.oob_prediction[index_join_oob] / self.oob_count[index_join_oob][:, None]
        elif self.is_ci:
            pred_time = self.oob_prediction[index_join_oob] / self.oob_count[index_join_oob]

        join_oob = self.y_train[index_join_oob]
        y_true = cnt.get_y(cens=join_oob[cnt.CENS_NAME], time=join_oob[cnt.TIME_NAME])
        return metr.METRIC_DICT[self.ens_metric_name.upper()](self.y_train, y_true, pred_time, pred_sf, pred_hf, self.bins)



import numpy as np

from .. import metrics as metr
from ..tree import CRAID
from .. import constants as cnt
from .boosting import BoostingCRAID
import matplotlib.pyplot as plt

from ..external import NelsonAalen, KaplanMeierZeroAfter


def f_predict_survival_at_times(ch, bins=None):
    mean, std, prob = ch
    durs = np.random.normal(mean, std, 1000)
    events = np.random.binomial(1, prob, size=1000)
    events = events[durs > 0]
    durs = durs[durs > 0]

    survival = KaplanMeierZeroAfter()
    survival.fit(durs, events)
    if bins is None:
        bins = survival.timeline
    sf = survival.survival_function_at_times(bins)
    return sf


def f_predict_hazard_at_times(ch, bins=None):
    mean, std, prob = ch
    durs = np.random.normal(mean, std, 1000)
    events = np.random.binomial(1, prob, size=1000)
    events = events[durs > 0]
    durs = durs[durs > 0]

    hazard = NelsonAalen()
    hazard.fit(durs, events)
    if bins is None:
        bins = hazard.timeline
    hf = hazard.cumulative_hazard_at_times(bins)
    return hf

class IBSCleverBoostingCRAID(BoostingCRAID):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IBSCleverBoostingCRAID"

    def fit(self, X, y):
        self.features = X.columns
        X = X.reset_index(drop=True)
        X[cnt.CENS_NAME] = y[cnt.CENS_NAME].astype(np.int32)
        X[cnt.TIME_NAME] = y[cnt.TIME_NAME].astype(np.float32)

        self.X_train = X
        self.X_train["ind_start"] = self.X_train.index
        self.y_train = y

        self.weights = np.ones(self.X_train.shape[0], dtype=float)
        self.bettas = []
        self.l_ibs = []
        self.l_weights = []
        self.update_params()

        for i in range(self.n_estimators):
            x_sub = self.X_train.sample(n=self.size_sample,
                                        # weights=self.weights,
                                        replace=self.bootstrap, random_state=i)

            x_oob = self.X_train.loc[self.X_train.index.difference(x_sub.index), :]
            print(f"UNIQUE ({i}):{np.unique(x_sub.index).shape[0]}, DIST:", np.bincount(x_sub["cens"]))
            x_sub = x_sub.reset_index(drop=True)
            X_sub_tr, y_sub_tr = cnt.pd_to_xy(x_sub)
            if self.weighted_tree:
                X_sub_tr["weights_obs"] = self.weights[x_sub['ind_start']]

            model = CRAID(features=self.features, random_state=i, **self.tree_kwargs)
            model.fit(X_sub_tr, y_sub_tr)

            wei_i, betta_i = self.count_model_weights(model, X_sub_tr, y_sub_tr)
            self.add_model(model, x_oob, wei_i, betta_i)
            self.update_weight(x_sub['ind_start'], wei_i)

    def predict(self, x_test, aggreg=True, **kwargs):
        res = []
        weights = []
        for i in range(len(self.models)):
            res.append(self.models[i].predict(x_test, **kwargs))

        res = np.array(res)
        weights = None
        if aggreg:
            res = self.get_aggreg(res, weights)
        return res

    def predict_at_times(self, x_test, bins, aggreg=True, mode="surv"):
        res = []
        weights = []
        for i in range(len(self.models)):
            res.append(self.models[i].predict_at_times(x_test, bins=bins,
                                                       mode=mode))

        res = np.array(res)
        weights = None
        if aggreg:
            res = self.get_aggreg(res, weights)
            if mode == "surv":
                res[:, -1] = 0
                res[:, 0] = 1
        return res

    # def predict_at_times(self, x_test, bins, aggreg=True, mode="surv"):
    #     res = []
    #     for i in range(len(self.models)):
    #         res.append(self.models[i].predict(x_test, target="ch")[np.newaxis, :])
    #
    #     res = np.vstack(res)
    #     means, stds, probs = res.T
    #     mean = np.mean(means, axis=1)
    #     std = np.sqrt(np.sum(stds ** 2, axis=1)) / np.sqrt(stds.shape[1])
    #     # std = np.sqrt(np.sum(stds ** 2, axis=1)) / np.sqrt(stds.shape[1])
    #     #         std = np.sqrt(np.sum((1/stds.shape[1]*stds)**2, axis=1))*1/np.sqrt(stds.shape[1])
    #     #         std = np.mean(stds, axis=1)
    #     prob = np.mean(probs, axis=1)
    #
    #     prepared = np.vstack([mean, std, prob]).T
    #     if mode == "surv":
    #         res = np.array(list(map(lambda x: f_predict_survival_at_times(x, bins), prepared)))
    #         res[:, -1] = 0
    #         res[:, 0] = 1
    #     else:
    #         res = np.array(list(map(lambda x: f_predict_hazard_at_times(x, bins), prepared)))
    #     return res

    def count_model_weights(self, model, X_sub, y_sub):
        # if self.all_weight:
        #     X_sub = self.X_train
        #     y_sub = self.y_train
        # pred_sf = model.predict_at_times(X_sub, bins=self.bins, mode="surv")
        #
        # ibs_sf = metr.auprc(self.y_train, y_sub, pred_sf, self.bins, axis=0)
        # betta = np.mean(ibs_sf)
        # wei = 1 - ibs_sf
        wei = np.ones(X_sub.shape[0])
        betta = 1
        return wei, abs(betta)

    def update_weight(self, index, wei_i):
        # self.weights += wei_i
        pass

    def get_aggreg(self, x, wei=None):
        if self.aggreg_func == 'median':
            return np.median(x, axis=0)
        elif self.aggreg_func == "wei":
            if wei is None:
                wei = np.array(self.bettas)
            wei = wei / np.sum(wei)
            return np.sum((x.T * wei).T, axis=0)
        elif self.aggreg_func == "argmean":
            wei = np.where(np.argsort(np.argsort(wei, axis=1), axis=1) > len(self.bettas) // 2, 1, 0)
            wei = wei / np.sum(wei, axis=1).reshape(-1, 1)
            return np.sum((x.T * wei).T, axis=0)
        elif self.aggreg_func == "argwei":
            wei = np.where(np.argsort(np.argsort(wei, axis=1), axis=1) > len(self.bettas) // 2,
                           1 / np.array(self.bettas), 0)
            wei = wei / np.sum(wei, axis=1).reshape(-1, 1)
            return np.sum((x.T * wei).T, axis=0)
        return np.mean(x, axis=0)

    def plot_curve(self, X_tmp, y_tmp, bins, label="", metric="ibs"):
        res = []
        metr_vals = []
        for i in range(len(self.models)):
            res.append(self.models[i].predict_at_times(X_tmp, bins=bins, mode="surv"))

            res_all = np.array(res)
            res_all = self.get_aggreg(res_all, np.array(self.bettas)[:i + 1])
            res_all[:, -1] = 0
            res_all[:, 0] = 1
            if metric == "ibs":
                metr_vals.append(metr.ibs_WW(self.y_train, y_tmp, res_all, bins))
            else:
                metr_vals.append(metr.auprc(self.y_train, y_tmp, res_all, bins))
        plt.plot(range(len(self.models)), metr_vals, label=label)

# class IBSCleverBoostingCRAID(BoostingCRAID):
#     def __init__(self, **kwargs):
#         super().__init__(**kwargs)
#         self.name = "IBSCleverBoostingCRAID"
#
#     def fit(self, X, y):
#         self.features = list(X.columns)
#         X = X.reset_index(drop=True).copy()
#         X[cnt.CENS_NAME] = y[cnt.CENS_NAME].astype(np.int32)
#         X[cnt.TIME_NAME] = y[cnt.TIME_NAME].astype(np.float32)
#
#         self.X_train = X.copy()
#         self.X_train["ind_start"] = self.X_train.index
#         self.y_train = y.copy()
#
#         self.weights = y[cnt.TIME_NAME].copy()
#         self.bettas = []
#         self.l_ibs = []
#         self.l_weights = []
#         self.update_params()
#         self.lr = 1
#
#         for i in range(self.n_estimators):
#             x_sub = self.X_train.sample(n=self.size_sample,
#                                         # weights=self.weights,
#                                         replace=self.bootstrap, random_state=i)
#
#             x_oob = self.X_train.loc[self.X_train.index.difference(x_sub.index), :]
#             # print(f"UNIQUE ({i}):{np.unique(x_sub.index).shape[0]}, DIST:", np.bincount(x_sub["cens"]))
#             x_sub = x_sub.reset_index(drop=True)
#             X_sub_tr, y_sub_tr = cnt.pd_to_xy(x_sub.copy())
#             y_sub_tr[cnt.TIME_NAME] = self.weights[x_sub['ind_start'].values].copy()
#
#             # plt.scatter(self.y_train["time"], self.weights, c=np.where(self.y_train["cens"], "orange", "blue"))
#             # plt.show()
#
#             model = CRAID(features=self.features.copy(), random_state=i, **self.tree_kwargs)
#             model.fit(X_sub_tr, y_sub_tr)
#
#             wei_i, betta_i = self.count_model_weights(model, X_sub_tr.copy(), y_sub_tr.copy())
#             self.add_model(model, x_oob, wei_i, betta_i)
#             self.update_weight(x_sub['ind_start'].values, wei_i)
#
#     def predict(self, x_test, aggreg=True, **kwargs):
#         res = []
#         weights = []
#         for i in range(len(self.models)):
#             res.append(self.models[i].predict(x_test, **kwargs))
#
#         res = np.array(res)
#         weights = None
#         if aggreg:
#             res = self.get_aggreg(res, weights)
#         return res
#
#     def predict_at_times(self, x_test, bins, aggreg=True, mode="surv"):
#         res = []
#         weights = []
#         for i in range(len(self.models)):
#             res.append(self.models[i].predict_at_times(x_test, bins=bins,
#                                                        mode=mode))
#
#         res = np.array(res)
#         weights = None
#         if aggreg:
#             res = self.get_aggreg(res, weights)
#             if mode == "surv":
#                 res[:, -1] = 0
#                 res[:, 0] = 1
#         return res
#
#     def count_model_weights(self, model, X_sub, y_sub):
#         if self.all_weight:
#             X_sub = self.X_train.copy()
#             y_sub = self.y_train.copy()
#
#         y_sub["time"] = self.weights[X_sub['ind_start'].values]
#         # pred_time = model.predict(X_sub, target="time")
#         pred_sf = model.predict_at_times(X_sub, bins=self.bins, mode="surv")
#
#         pred_time = self.bins[(pred_sf <= 0.5).argmax(axis=1)]
#         ibs_sf = metr.ibs_remain(self.y_train, y_sub, pred_sf, self.bins, axis=0)
#
#         wei = (y_sub["time"] - pred_time) * ibs_sf**2  # (2*y_sub["cens"] - 1) *
#         betta = self.lr / np.mean(ibs_sf)
#         return wei, abs(betta)
#
#     def update_weight(self, index, wei_i):
#         #         print(index, wei_i)
#         self.lr = 0.3
#         if self.all_weight:
#             self.weights -= wei_i * self.lr
#         else:
#             self.weights[index] -= wei_i * self.lr
#
#     def get_aggreg(self, x, wei=None):
#         if self.aggreg_func == 'median':
#             return np.median(x, axis=0)
#         elif self.aggreg_func == "wei":
#             if wei is None:
#                 wei = np.array(self.bettas)
#             wei = wei / np.sum(wei)
#             return np.sum((x.T * wei).T, axis=0)
#         elif self.aggreg_func == "argmean":
#             wei = np.where(np.argsort(np.argsort(wei, axis=1), axis=1) > len(self.bettas) // 2, 1, 0)
#             wei = wei / np.sum(wei, axis=1).reshape(-1, 1)
#             return np.sum((x.T * wei).T, axis=0)
#         elif self.aggreg_func == "argwei":
#             wei = np.where(np.argsort(np.argsort(wei, axis=1), axis=1) > len(self.bettas) // 2,
#                            1 / np.array(self.bettas), 0)
#             wei = wei / np.sum(wei, axis=1).reshape(-1, 1)
#             return np.sum((x.T * wei).T, axis=0)
#         return np.mean(x, axis=0)
#
#     def plot_curve(self, X_tmp, y_tmp, bins, label="", metric="ibs"):
#         res = []
#         metr_vals = []
#         for i in range(len(self.models)):
#             res.append(self.models[i].predict_at_times(X_tmp, bins=bins, mode="surv"))
#
#             res_all = np.array(res)
#             res_all = self.get_aggreg(res_all, np.array(self.bettas)[:i + 1])
#             res_all[:, -1] = 0
#             res_all[:, 0] = 1
#             if metric == "ibs":
#                 metr_vals.append(metr.ibs_WW(self.y_train, y_tmp, res_all, bins))
#             else:
#                 metr_vals.append(metr.auprc(self.y_train, y_tmp, res_all, bins))
#         plt.plot(range(len(self.models)), metr_vals, label=label)
#
#     def tolerance_find_best(self, ens_metric_name="bic"):
#         pass



from .other import load_gbsg_dataset
from .other import load_pbc_dataset
from .other import load_wuhan_dataset

from .other import load_actg_dataset
from .other import load_flchain_dataset
from .other import load_smarto_dataset
from .other import load_rott2_dataset
from .other import load_support2_dataset
from .other import load_Framingham_dataset

from .covid import load_covid_dataset
from .onk import load_onk_dataset
from .backblaze import load_backblaze_dataset

from .new_backblaze import load_backblaze_2016_2018
from .new_backblaze import load_backblaze_2018_2021
from .new_backblaze import load_backblaze_2021_2023

DATASETS_LOAD = {
    "GBSG": load_gbsg_dataset,
    "PBC": load_pbc_dataset,
    "WUHAN": load_wuhan_dataset,
    "actg": load_actg_dataset,
    "flchain": load_flchain_dataset,
    "smarto": load_smarto_dataset,
    "rott2": load_rott2_dataset,
    "support2": load_support2_dataset,
    "Framingham": load_Framingham_dataset,
    "backblaze16_18": load_backblaze_2016_2018,
    "backblaze18_21": load_backblaze_2018_2021,
    "backblaze21_23": load_backblaze_2021_2023,
    # "backblaze": load_backblaze_dataset
    # "ONK": load_onk_dataset,
    # "COVID": load_covid_dataset,
}



from .leaf_model import LeafModel, NonparamLeafModel

from .parametric import WeibullAFT, LogNormalAFT, LogLogisticAFT, AFT_param_grid
from .parametric import CoxPH, CoxPH_param_grid
from .parametric import LEAF_AFT_DICT

from .nonparametric import KaplanMeier, KaplanMeierZeroAfter, FullProbKM
from .nonparametric import NelsonAalen
from .nonparametric import LEAF_NONPARAM_DICT

LEAF_MODEL_DICT = LEAF_NONPARAM_DICT.copy()
LEAF_MODEL_DICT.update(LEAF_AFT_DICT)



from sksurv.metrics import check_y_survival, CensoringDistributionEstimator
import numpy as np
from numba import njit, jit
from lifelines import KaplanMeierFitter, NelsonAalenFitter
# from lifelines.utils import concordance_index

from .constants import TIME_NAME, CENS_NAME

METRIC_DICT = {
    "CI": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        concordance_index(y_tst[TIME_NAME], pred_time),
    "CI_CENS": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        concordance_index(y_tst[TIME_NAME], pred_time, y_tst[CENS_NAME]),

    "IBS": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        ibs(y_tr, y_tst, pred_sf, bins),
    "BAL_IBS": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        bal_ibs(y_tr, y_tst, pred_sf, bins),

    "IBS_WW": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        ibs_WW(y_tr, y_tst, pred_sf, bins),
    "BAL_IBS_WW": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        bal_ibs_WW(y_tr, y_tst, pred_sf, bins),

    "IBS_REMAIN": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        ibs_remain(y_tr, y_tst, pred_sf, bins),
    "BAL_IBS_REMAIN": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        bal_ibs_remain(y_tr, y_tst, pred_sf, bins),

    "IAUC": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        iauc(y_tr, y_tst, pred_hf, bins),
    "IAUC_WW": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        iauc_WW(y_tr, y_tst, pred_hf, bins),
    "IAUC_TI": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        iauc_TI(y_tr, y_tst, pred_hf, bins),
    "IAUC_WW_TI": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        iauc_WW_TI(y_tr, y_tst, pred_hf, bins),

    "AUPRC": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        auprc(y_tr, y_tst, pred_sf, bins),
    "AUPRC_by_obs": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        auprc(y_tr, y_tst, pred_sf, bins, axis=3),
    "EVENT_AUPRC": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        event_auprc(y_tr, y_tst, pred_sf, bins),
    "CENS_AUPRC": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        cens_auprc(y_tr, y_tst, pred_sf, bins),
    "BAL_AUPRC": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        bal_auprc(y_tr, y_tst, pred_sf, bins),

    "LOGLIKELIHOOD": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        loglikelihood(y_tst[TIME_NAME], y_tst[CENS_NAME], pred_sf, pred_hf, bins),
    "KL": lambda y_tr, y_tst, pred_time, pred_sf, pred_hf, bins:
        kl(y_tst[TIME_NAME], y_tst[CENS_NAME], pred_sf, pred_hf, bins)
}
""" dict: Available metrics in library and its realization """

DESCEND_METRICS = ['ibs', 'IBS', 'aic', "AIC", "bic", "BIC", "KL",
                   "BAL_IBS", "IBS_WW", "BAL_IBS_WW", "IBS_REMAIN", "BAL_IBS_REMAIN"]
""" list: Metrics with decreasing quality improvement """


def auprc(survival_train, survival_test, estimate, times, axis=-1):
    time = survival_test["time"]
    event = survival_test["cens"]

    steps = np.linspace(1e-5, 1 - 1e-5, 100)
    before_time = np.dot(time[:, np.newaxis], steps[np.newaxis, :])
    after_time = np.dot(time[:, np.newaxis], 1 / steps[np.newaxis, :])  # TODO OLD

    before_ind = np.clip(np.searchsorted(times, before_time), 0, times.shape[0] - 1)
    after_ind = np.clip(np.searchsorted(times, after_time), 0, times.shape[0] - 1)

    est = np.take_along_axis(estimate, before_ind, axis=1)
    est[event] -= np.take_along_axis(estimate[event], after_ind[event], axis=1)

    if axis == -1:  # mean for each time and observation
        est = np.mean(est, axis=0)  # TODO np.mean
        return np.trapz(est, steps)
        # return np.median(np.trapz(est, steps))
    elif axis == 0:  # for each observation
        return np.trapz(est, steps)
    elif axis == 1:  # in time (for graphics)
        # est = est.median(axis=0)
        est = est.mean(axis=0)  # TODO np.mean
        return est
    elif axis == 2:  # source
        return est
    elif axis == 3:  # for each observation with array wrap
        return np.array([np.trapz(est, steps)])
    return None


def bal_auprc(survival_train, survival_test, estimate, times, axis=-1):
    auprc_event = auprc(survival_train, survival_test[survival_test["cens"]],
                        estimate[survival_test["cens"]], times, axis=axis)
    auprc_cens = auprc(survival_train, survival_test[~survival_test["cens"]],
                       estimate[~survival_test["cens"]], times, axis=axis)
    return (auprc_event + auprc_cens) / 2


def event_auprc(survival_train, survival_test, estimate, times, axis=-1):
    auprc_event = auprc(survival_train, survival_test[survival_test["cens"]],
                        estimate[survival_test["cens"]], times, axis=axis)
    return auprc_event


def cens_auprc(survival_train, survival_test, estimate, times, axis=-1):
    auprc_cens = auprc(survival_train, survival_test[~survival_test["cens"]],
                       estimate[~survival_test["cens"]], times, axis=axis)
    return auprc_cens


@njit
def get_before(estim, wei):
    return np.square(estim) * wei


@njit
def get_after(estim, prob_cens):
    return np.square(1 - estim) / prob_cens


def ibs(survival_train, survival_test, estimate, times, axis=-1):
    """
    Modified integrated brier score from scikit-survival (add axis)
    Modification: with axis = 0 count ibs for each observation

    Parameters
    ----------
    survival_train : structured array, shape = (n_train_samples,)
        Survival times for training data to estimate the censoring
        distribution from.
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.
    survival_test : structured array, shape = (n_samples,)
        Survival times of test data.
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.
    estimate : array-like, shape = (n_samples, n_times)
        Estimated risk of experiencing an event for test data at `times`.
        The i-th column must contain the estimated probability of
        remaining event-free up to the i-th time point.
    times : array-like, shape = (n_times,)
        The time points for which to estimate the Brier score.
    axis : int, optional
        With axis = 1 count bs for each time from times
        With axis = 0 count ibs for each observation.
        With axis = -1 count mean ibs.
        The default is -1.

    Returns
    -------
    ibs_value : float or array-like or None
        if axis = 0 return array of ibs for each observation
           axis = 1 return array of bs for each time from times
           axis = -1 return float
        else
            None

    """
    test_event, test_time = check_y_survival(survival_test, allow_all_censored=True)
    # estimate, times = _check_estimate_2d(estimate, test_time, times)
    estimate = np.array(estimate)
    if estimate.ndim == 1 and times.shape[0] == 1:
        estimate = estimate.reshape(-1, 1)
    estimate[estimate == -np.inf] = 0
    estimate[estimate == np.inf] = 0
    # fit IPCW estimator
    cens = CensoringDistributionEstimator().fit(survival_train)
    # calculate inverse probability of censoring weight at current time point t.
    prob_cens_t = cens.predict_proba(times)
    prob_cens_t[prob_cens_t == 0] = np.inf
    # calculate inverse probability of censoring weights at observed time point
    prob_cens_y = cens.predict_proba(test_time)
    prob_cens_y[prob_cens_y == 0] = np.inf
    wei = test_event / prob_cens_y
    
    estim_before = get_before(estimate, wei[np.newaxis, :].T)
    estim_after = get_after(estimate, prob_cens_t)
    brier_scores = np.array([np.where(test_time <= t, 
                                      estim_before[:, i], 
                                      estim_after[:, i])
                             for i, t in enumerate(times)])
    time_diff = times[-1] - times[0] if times[-1] > times[0] else 1
    if axis == -1:  # mean ibs for each time and observation
        brier_scores = np.mean(brier_scores, axis=1)  # TODO np.mean
        return np.trapz(brier_scores, times) / time_diff
        # return np.median(np.trapz(brier_scores, times, axis=0)) / time_diff
    elif axis == 0:  # ibs for each observation
        return np.trapz(brier_scores, times, axis=0) / time_diff
    elif axis == 1:  # bs in time (for graphics)
        # return np.median(brier_scores, axis=1)
        return np.mean(brier_scores, axis=1)  # TODO np.mean
    return None


def bal_ibs(survival_train, survival_test, estimate, times, axis=-1):
    """ IBS with equal impact of each event type """
    ibs_event = ibs(survival_train, survival_test[survival_test["cens"]],
                    estimate[survival_test["cens"]], times, axis=axis)
    ibs_cens = ibs(survival_train, survival_test[~survival_test["cens"]],
                   estimate[~survival_test["cens"]], times, axis=axis)
    return ibs_event + ibs_cens


def ibs_WW(survival_train, survival_test, estimate, times, axis=-1):
    """ IBS with equal impact of partial observation """
    test_event, test_time = check_y_survival(survival_test, allow_all_censored=True)
    estimate = np.array(estimate)
    if estimate.ndim == 1 and times.shape[0] == 1:
        estimate = estimate.reshape(-1, 1)
    estimate[estimate == -np.inf] = 0
    estimate[estimate == np.inf] = 0

    estim_before = np.square(estimate) * test_event[np.newaxis, :].T
    estim_after = np.square(1 - estimate)
    brier_scores = np.array([np.where(test_time <= t,
                                      estim_before[:, i],
                                      estim_after[:, i])
                             for i, t in enumerate(times)])
    time_diff = times[-1] - times[0] if times[-1] > times[0] else 1
    if axis == -1:  # mean ibs for each time and observation
        brier_scores = np.mean(brier_scores, axis=1)  # TODO np.mean
        return np.trapz(brier_scores, times) / time_diff
        # return np.median(np.trapz(brier_scores, times, axis=0)) / time_diff
    elif axis == 0:  # ibs for each observation
        return np.trapz(brier_scores, times, axis=0) / time_diff
    elif axis == 1:  # bs in time (for graphics)
        # return np.median(brier_scores, axis=1)
        return np.mean(brier_scores, axis=1)  # TODO np.mean
    return None


def bal_ibs_WW(survival_train, survival_test, estimate, times, axis=-1):
    """ IBS with equal impact of each event type and partial observation """
    ibs_event = ibs_WW(survival_train, survival_test[survival_test["cens"]],
                       estimate[survival_test["cens"]], times, axis=axis)
    ibs_cens = ibs_WW(survival_train, survival_test[~survival_test["cens"]],
                      estimate[~survival_test["cens"]], times, axis=axis)
    return ibs_event + ibs_cens


def ibs_remain(survival_train, survival_test, estimate, times, axis=-1):
    """ IBS with equal impact of partial observation with controlled quantity """
    test_event, test_time = check_y_survival(survival_test, allow_all_censored=True)
    estimate = np.array(estimate)
    if estimate.ndim == 1 and times.shape[0] == 1:
        estimate = estimate.reshape(-1, 1)
    estimate[estimate == -np.inf] = 0
    estimate[estimate == np.inf] = 0

    estim_before = np.square(estimate) * test_event[np.newaxis, :].T
    estim_after = np.square(1 - estimate)
    brier_scores = np.array([np.where(test_time < t,
                                      estim_before[:, i],
                                      estim_after[:, i])
                             for i, t in enumerate(times)])
    N = np.sum(np.array([np.where(test_time < t, test_event, 1)
                         for i, t in enumerate(times)]), axis=1)
    # ind = np.digitize(test_time, times)
    # n_cens = np.bincount(ind[~test_event], minlength=times.shape[0])
    #
    # N = np.ones(times.shape) * np.sum(test_event)
    # if n_cens.shape[0] > 0:
    #     N += np.cumsum(n_cens[::-1])[::-1]
    time_diff = times[-1] - times[0] if times[-1] > times[0] else 1
    if axis == -1:  # mean ibs for each time and observation
        # brier_scores = np.mean(brier_scores, axis=1)
        brier_scores = np.where(N > 0, 1 / N, 0) * np.sum(brier_scores, axis=1)
        return np.trapz(brier_scores, times) / time_diff
    elif axis == 0:  # ibs for each observation
        return np.trapz(brier_scores, times, axis=0) / time_diff
    elif axis == 1:  # bs in time (for graphics)
        # brier_scores = np.mean(brier_scores, axis=1)
        brier_scores = np.where(N > 0, 1 / N, 0) * np.sum(brier_scores, axis=1)
        return brier_scores
    return None


def bal_ibs_remain(survival_train, survival_test, estimate, times, axis=-1):
    """ IBS with equal impact of each event type and partial observation with controlled quantity """
    ibs_event = ibs_remain(survival_train, survival_test[survival_test["cens"]],
                           estimate[survival_test["cens"]], times, axis=axis)
    ibs_cens = ibs_remain(survival_train, survival_test[~survival_test["cens"]],
                          estimate[~survival_test["cens"]], times, axis=axis)
    return ibs_event + ibs_cens


def iauc(survival_train, survival_test, estimate, times, tied_tol=1e-8, no_wei=False, time_int=False):
    """
    Modified integrated AUC (cumulative_dynamic_auc) 
        from scikit-survival (reduce complexity)

    Parameters
    ----------
    survival_train : structured array, shape = (n_train_samples,)
        Survival times for training data to estimate the censoring
        distribution from.
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.
    survival_test : structured array, shape = (n_samples,)
        Survival times of test data.
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.
    estimate : array-like, shape = (n_samples, n_times)
        Estimated risk of experiencing an event for test data at `times`.
        The i-th column must contain the estimated probability of
        remaining event-free up to the i-th time point.
    times : array-like, shape = (n_times,)
        The time points for which to estimate the Brier score.
    tied_tol : float, optional, default: 1e-8
        The tolerance value for considering ties.
        If the absolute difference between risk scores is smaller
        or equal than `tied_tol`, risk scores are considered tied.
        
    Returns
    -------
    mean_auc : float
        Summary measure referring to the mean cumulative/dynamic AUC
        over the specified time range `(times[0], times[-1])`.

    """
    if survival_train[CENS_NAME].sum() == 0:
        survival_train[CENS_NAME] = 1
        survival_test[CENS_NAME] = 1 - survival_test[CENS_NAME]
    if survival_test[CENS_NAME].sum() == 0:
        survival_test[CENS_NAME] = 1
    test_event, test_time = check_y_survival(survival_test)
    # estimate, times = _check_estimate_2d(estimate, test_time, times)
    estimate = np.array(estimate)
    n_samples = estimate.shape[0]
    n_times = times.shape[0]
    if estimate.ndim == 1:
        estimate = np.broadcast_to(estimate[:, np.newaxis], (n_samples, n_times))

    # fit and transform IPCW
    if no_wei:
        ipcw = np.ones(test_time.shape[0])
    else:
        cens = CensoringDistributionEstimator()
        cens.fit(survival_train)
        g_hat = cens.predict_proba(test_time[test_event])
        ipcw = np.zeros(test_time.shape[0])
        g_hat[g_hat == 0] = np.inf
        if not ((g_hat == 0.0).any()):
            ipcw[test_event] = 1.0 / g_hat
        else:
            ipcw = np.ones(test_time.shape[0])

    # expand arrays to (n_samples, n_times) shape
    test_time = np.broadcast_to(test_time[:, np.newaxis], (n_samples, n_times))
    test_event = np.broadcast_to(test_event[:, np.newaxis], (n_samples, n_times))
    times_2d = np.broadcast_to(times, (n_samples, n_times))
    ipcw = np.broadcast_to(ipcw[:, np.newaxis], (n_samples, n_times))

    # sort each time point (columns) by risk score (descending)
    o = np.argsort(-estimate, axis=0)
    test_time = np.take_along_axis(test_time, o, axis=0)
    test_event = np.take_along_axis(test_event, o, axis=0)
    estimate = np.take_along_axis(estimate, o, axis=0)
    ipcw = np.take_along_axis(ipcw, o, axis=0)

    is_case = (test_time <= times_2d) & test_event
    is_control = test_time > times_2d
    n_controls = is_control.sum(axis=0)
    n_controls[n_controls == 0] = 1

    # prepend row of infinity values
    estimate_diff = np.concatenate((np.broadcast_to(np.infty, (1, n_times)), estimate))
    is_tied = np.absolute(np.diff(estimate_diff, axis=0)) <= tied_tol

    cumsum_tp = np.cumsum(is_case * ipcw, axis=0)
    cumsum_fp = np.cumsum(is_control, axis=0)
    true_pos = cumsum_tp / cumsum_tp[-1]
    false_pos = cumsum_fp / n_controls

    scores = np.empty(n_times, dtype=float)
    it = np.nditer((true_pos, false_pos, is_tied), order="F", flags=["external_loop"])
    with it:
        for i, (tp, fp, mask) in enumerate(it):
            idx = np.flatnonzero(mask) - 1
            # only keep the last estimate for tied risk scores
            tp_no_ties = np.delete(tp, idx)
            fp_no_ties = np.delete(fp, idx)
            # Add an extra threshold position
            # to make sure that the curve starts at (0, 0)
            tp_no_ties = np.r_[0, tp_no_ties]
            fp_no_ties = np.r_[0, fp_no_ties]
            scores[i] = np.trapz(tp_no_ties, fp_no_ties)

    scores[np.isnan(scores)] = 0
    if n_times != 1:
        if time_int:
            time_diff = times[-1] - times[0] if times[-1] > times[0] else 1
            return np.trapz(scores, times, axis=0) / time_diff
        else:
            surv = KaplanMeierFitter()
            surv.fit(survival_test[TIME_NAME], survival_test[CENS_NAME])
            s_times = surv.survival_function_at_times(times).to_numpy()

            # compute integral of AUC over survival function
            d = -np.diff(np.r_[1.0, s_times])
            integral = (scores * d).sum()
            return integral / (1.0 - s_times[-1])
    return scores[0]


def iauc_WW(s_tr, s_tst, est, times, tied_tol=1e-8):
    """ IAUC without weighting for each observation """
    return iauc(s_tr, s_tst, est, times, tied_tol=tied_tol, no_wei=True)


def iauc_TI(s_tr, s_tst, est, times, tied_tol=1e-8):
    """ IAUC with integration by time (instead of S(t)) """
    return iauc(s_tr, s_tst, est, times, tied_tol=tied_tol, no_wei=False, time_int=True)


def iauc_WW_TI(s_tr, s_tst, est, times, tied_tol=1e-8):
    """ IAUC with equal weight for each observation and integration by time """
    return iauc(s_tr, s_tst, est, times, tied_tol=tied_tol, no_wei=True, time_int=True)


def ipa(survival_train, survival_test, estimate, times, axis=-1):
    """
    Index of Prediction Accuracy: General R^2 for binary outcome and right
    censored time to event (survival) outcome also with competing risks.

    Parameters
    ----------
    survival_train : structured array, shape = (n_train_samples,)
        Survival times for training data to estimate the censoring
        distribution from.
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.
    survival_test : structured array, shape = (n_samples,)
        Survival times of test data.
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.
    estimate : array-like, shape = (n_samples, n_times)
        Estimated risk of experiencing an event for test data at `times`.
        The i-th column must contain the estimated probability of
        remaining event-free up to the i-th time point.
    times : array-like, shape = (n_times,)
        The time points for which to estimate the Brier score.
    axis : int, optional
        With axis = 1 count ipa for each time from times
        With axis = 0 count ipa for each observation.
        With axis = -1 count mean ipa.
        The default is -1.

    Returns
    -------
    ibs_value : float or array-like or None
        if axis = 0 return array
           axis = 1 return array
           axis = -1 return float
        else
            None

    """
    one_sf = get_survival_func(survival_train['time'], survival_train['cens'], times)[np.newaxis, :]
    kmf_estimate = np.repeat(one_sf, survival_test.shape[0], axis=0)

    ibs_model = ibs(survival_train, survival_test, estimate, times, axis)
    ibs_kmf_model = ibs(survival_train, survival_test, kmf_estimate, times, axis)
    return 1 - (ibs_model + 1e-8) / (ibs_kmf_model + 1e-8)


def kl(time, cens, sf, cumhf, bins):
    """ Kullback-Leibler divergence in terms of survival analysis """
    eq_sf = np.array([np.where(time > t, 1, 0)
                      for i, t in enumerate(bins)]).T

    kl_v = np.sum(sf * np.log((sf + 1e-10) / (eq_sf + 1e-10)) + np.abs(sf - eq_sf), axis=0)
    return np.trapz(kl_v, bins, axis=0) / (bins[-1] - bins[0])


def loglikelihood(time, cens, sf, cumhf, bins):
    """ Likelihood in terms of survival analysis (without PH) """
    index_times = np.digitize(time, bins, right=True) - 1
    hf = np.hstack((cumhf[:, 0][np.newaxis].T, np.diff(cumhf)))
    sf_by_times = np.take_along_axis(sf, index_times[:, np.newaxis], axis=1)[:, 0] + 1e-10
    hf_by_times = (np.take_along_axis(hf, index_times[:, np.newaxis], axis=1)[:, 0] + 1e-10)**cens
    likelihood = np.sum(np.log(sf_by_times) + np.log(hf_by_times))
    return likelihood


def aic(num_params, time, cens, sf, cumhf, bins):
    return 2*num_params - 2*loglikelihood(time, cens, sf, cumhf, bins)


def bic(k, n, time, cens, sf, cumhf, bins):
    return k*np.log(n) - 2*loglikelihood(time, cens, sf, cumhf, bins)


@njit
def count_pairs(T, P, E):
    n = len(T)
    concordant_pairs = 0
    total_pairs = 0
    for i in range(n):
        for j in range(i + 1, n):
            if E[i] == 1 and T[i] <= T[j]:
                total_pairs += 1
                concordant_pairs += P[i] < P[j]
                concordant_pairs += 0.5 * (P[i] == P[j])
    return concordant_pairs, total_pairs


def concordance_index(event_times, predicted_scores, event_observed=None):
    """
    Calculates the concordance index (C-index) for survival analysis.
    Previous speed: 684 ms ± 4.06 ms per loop (mean ± std. dev. of 7 runs, 1 loop each).
    Current speed: 22.7 ms ± 128 µs per loop (mean ± std. dev. of 7 runs, 10 loops each).
    Speed has increased by ~30 times

    Parameters
    ----------
    event_times: Array of true event times.
    predicted_scores: Array of predicted event times.
    event_observed: Array of event indicators (1 if event occurred, 0 if censored).

    Returns
    -------
    The concordance index: float

    Examples
    --------
    .. code:: python

        from survivors.metrics import concordance_index
        concordance_index(np.array([10, 20, 30, 40]),
                          np.array([20, 19, 29, 39]),
                          np.array([1, 0, 1, 0]))
    """
    if event_observed is None:
        event_observed = np.ones(len(event_times))
    order = np.argsort(event_times)
    predicted_scores = np.asarray(predicted_scores)[order]
    event_times = np.asarray(event_times)[order]
    event_observed = np.asarray(event_observed)[order]
    concordant_pairs, total_pairs = count_pairs(event_times, predicted_scores, event_observed)

    if total_pairs == 0:
        return 0
    return concordant_pairs / total_pairs


""" ESTIMATE FUNCTION """


def get_survival_func(ddeath, cdeath, bins=None):
    """
    Build Kaplan-Meier Estimate of survival function

    Parameters
    ----------
    ddeath : array-like
        Times of occurred events
    cdeath : array-like
        Indicate of occurred events (Censoring flag)
    bins : array-like, optional
        Points of survival function. The default is None.

    Returns
    -------
    KaplanMeierFitter or array
        If bins is None return kaplan-meier model, 
                   else return values of SF.

    """
    kmf = KaplanMeierFitter()
    kmf.fit(ddeath, cdeath)
    if not (bins is None):
        return kmf.survival_function_at_times(bins).to_numpy()
    return kmf


def get_hazard_func(ddeath, cdeath, bins=None):
    """
    Build Nelson-Aalen Estimate of Hazard function

    Parameters
    ----------
    ddeath : array-like
        Times of occurred events
    cdeath : array-like
        Indicate of occurred events (Censoring flag)
    bins : array-like, optional
        Points of hazard function. The default is None.

    Returns
    -------
    NelsonAalenFitter or array
        If bins is None return Nelson-Aalen model, 
                   else return values of HF.

    """
    naf = NelsonAalenFitter()
    naf.fit(ddeath, cdeath)
    if not (bins is None):
        return naf.cumulative_hazard_at_times(bins).to_numpy()
    return naf


IBS_DICT = {
    m.__name__.upper(): m
    for m in [ibs, bal_ibs,
              ibs_WW, bal_ibs_WW,
              ibs_remain, bal_ibs_remain]
}



import numpy as np
from scipy.stats import norm
from .leaf_model import NonparamLeafModel, MixLeafModel, NormalizedLeafModel, MeaningLeafModel, StableLeafModel
from lifelines import KaplanMeierFitter, NelsonAalenFitter


def epanechnikov_kernel(t, T, bandwidth=1.0):
    M = 0.75 * (1 - ((t - T) / bandwidth) ** 2)
    M[abs((t - T)) >= bandwidth] = 0
    return M


class KaplanMeier:
    """
    Modification of classic Kaplan-Meier model

    Attributes
    ----------
    timeline : numpy ndarray
        Bins of event time
    survival_function : numpy ndarray
        Survival probability for all points of timeline
    cumul_hist_dur : numpy ndarray
        Cumulative hazard probability for all points of timeline

    Methods
    -------
    fit : Build nonparametric model
    get_confidence_interval_ : predict confidence interval by times
    survival_function_at_times : predict survival probability by times
    """
    def __init__(self, **kwargs):
        self.timeline = None
        self.survival_function = None
        self.confidence_interval_ = None
        self.alpha = 0.05

    def fit(self, durations, right_censor, weights=None):
        durations = np.array(durations, dtype=float)
        right_censor = np.array(right_censor, dtype=int)
        if weights is None:
            weights = np.ones(right_censor.shape)
        self.timeline = np.unique(durations)

        dur_ = np.searchsorted(self.timeline, durations)
        hist_dur = np.bincount(dur_, weights=weights)
        self.hist_cens = np.bincount(dur_, weights=right_censor * weights)
        self.cumul_hist_dur = np.cumsum(hist_dur[::-1])[::-1]
        self.survival_function = np.hstack([1.0, np.cumprod((1.0 - self.hist_cens / (self.cumul_hist_dur)))])

    def count_confidence_interval(self):
        """ Calculated by exponential Greenwood: https://www.math.wustl.edu/~sawyer/handouts/greenwood.pdf """
        z = norm.ppf(1 - self.alpha / 2)
        cumulative_sq_ = np.sqrt(np.hstack(
            [0.0, np.cumsum(self.hist_cens / (self.cumul_hist_dur * (self.cumul_hist_dur - self.hist_cens)))]))
        np.nan_to_num(cumulative_sq_, copy=False, nan=0)
        v = np.log(self.survival_function)
        np.nan_to_num(v, copy=False, nan=0)
        self.confidence_interval_ = np.vstack([np.exp(v * np.exp(- z * cumulative_sq_ / v)),
                                               np.exp(v * np.exp(+ z * cumulative_sq_ / v))]).T
        np.nan_to_num(self.confidence_interval_, copy=False, nan=1)

    def get_confidence_interval_(self):
        if self.confidence_interval_ is None:
            self.count_confidence_interval()
        return self.confidence_interval_

    def survival_function_at_times(self, times):
        place_bin = np.digitize(times, self.timeline)
        return self.survival_function[np.clip(place_bin, 0, None)]


class FullProbKM(KaplanMeier):
    """
    A KM model predicting the survival function using the full probability formula
    """
    def fit(self, durations, right_censor, weights=None):
        durations = np.array(durations, dtype=float)
        right_censor = np.array(right_censor, dtype=int)
        if weights is None:
            weights = np.ones(right_censor.shape)

        self.timeline = np.unique(durations)
        right_censor = right_censor.astype("bool")
        dur_ = np.searchsorted(self.timeline, durations)

        self.hist_cens = np.bincount(dur_, weights=right_censor * weights)
        self.cumul_hist_dur = np.cumsum(self.hist_cens[::-1])[::-1]
        self.cumul_hist_dur[self.cumul_hist_dur == 0] = 1e-3  # Any cnt (in sf it becomes zero)

        self.survival_function = np.hstack([1.0, np.cumprod((1.0 - self.hist_cens / self.cumul_hist_dur))])

        N = right_censor.shape[0]
        Ncens = right_censor[~right_censor].shape[0]
        self.survival_function = Ncens / N + (1 - Ncens / N) * self.survival_function


class NelsonAalen:
    """
    Modification of classic Nelson-Aalen model

    Attributes
    ----------
    timeline : numpy ndarray
        Bins of event time
    survival_function : numpy ndarray
        Survival probability for all points of timeline
    smoothing : bool
        Flag the need for density smoothing

    Methods
    -------
    fit : Build nonparametric model
    cumulative_hazard_at_times : predict hazard probability by times
    get_smoothed_hazard_at_times : predict smoothed hazard probability by times
    """
    def __init__(self, smoothing=True, **kwargs):
        self.timeline = None
        self.survival_function = None
        self.smoothing = smoothing

    def fit(self, durations, right_censor, weights=None):
        durations = np.array(durations, dtype=float)
        right_censor = np.array(right_censor, dtype=int)
        if weights is None:
            weights = np.ones(right_censor.shape)
        # The formula Stata: https://stats.stackexchange.com/questions/6670/
        self.bandwidth = np.std(durations) / (len(durations) ** (1 / 5))
        self.timeline = np.unique(durations)

        dur_ = np.searchsorted(self.timeline, durations)
        hist_dur = np.bincount(dur_, weights=weights)
        hist_cens = np.bincount(dur_, weights=right_censor * weights)
        cumul_hist_dur = np.cumsum(hist_dur[::-1])[::-1]
        if self.smoothing and all(weights == 1):
            cumul_hist_dur = cumul_hist_dur.astype("int")
            hist_cens = hist_cens.astype("int")
            cum_ = np.cumsum(1.0 / np.arange(1, np.max(cumul_hist_dur) + 1))
            hf = cum_[cumul_hist_dur - 1] - np.where(cumul_hist_dur - hist_cens - 1 >= 0,
                                                     cum_[cumul_hist_dur - hist_cens - 1], 0)
        else:
            hf = hist_cens / cumul_hist_dur
        self.hazard_function = np.hstack([0.0, np.cumsum(hf)])

    def cumulative_hazard_at_times(self, times):
        place_bin = np.digitize(times, self.timeline)
        return self.hazard_function[np.clip(place_bin, 0, None)]

    def smoothed_hazard_(self, bandwidth):
        timeline = self.timeline
        hazard_ = np.diff(self.hazard_function)
        sh = 1.0 / bandwidth * np.dot(epanechnikov_kernel(timeline[:, None],
                                                          timeline[None, :],
                                                          bandwidth), hazard_)
        return sh + np.max(sh) / self.timeline.shape[0]

    def get_smoothed_hazard_at_times(self, bins):
        hazard_ = np.hstack([0.0, np.diff(self.cumulative_hazard_at_times(bins))])
        sh = 1.0 / self.bandwidth * np.dot(epanechnikov_kernel(bins[:, None],
                                                               bins[None, :],
                                                               self.bandwidth), hazard_)
        return sh + np.max(sh) / bins.shape[0]


class KaplanMeierZeroAfter(KaplanMeier):
    """
    External pre-definition of the left and right KM estimation bounds with one and zero.
    """
    def survival_function_at_times(self, times):
        place_bin = np.searchsorted(self.timeline, times)
        # place_bin = np.digitize(times, self.timeline)  # -1
        sf = self.survival_function[np.clip(place_bin, 0, None)]
        sf[times > self.timeline[-1]] = 0
        sf[times < self.timeline[0]] = 1
        return sf


class BaseLeafModel(NonparamLeafModel):
    survival_class = KaplanMeier
    hazard_class = NelsonAalen


class BaseLeafModelOnlySurv(NonparamLeafModel):
    survival_class = KaplanMeier


class BaseLeafModelOnlyHazard(NonparamLeafModel):
    hazard_class = NelsonAalen


class BaseNormalizedLeafModel(NormalizedLeafModel):
    survival_class = KaplanMeierZeroAfter
    hazard_class = NelsonAalen


class StableLeafModel(StableLeafModel):
    survival_class = KaplanMeierZeroAfter
    hazard_class = NelsonAalen


class BaseMeaningLeafModel(MeaningLeafModel):
    survival_class = KaplanMeierZeroAfter
    hazard_class = NelsonAalen


class BaseMixLeafModel(MixLeafModel):
    survival_class = KaplanMeierZeroAfter
    hazard_class = NelsonAalen


class BaseLeafModeLL(NonparamLeafModel):
    survival_class = KaplanMeierFitter
    hazard_class = NelsonAalenFitter


LEAF_NONPARAM_DICT = {
    "base": BaseLeafModel,
    "baseLL": BaseLeafModeLL,
    "only_hazard": BaseLeafModelOnlyHazard,
    "only_survive": BaseLeafModelOnlySurv,
    "base_zero_after": BaseNormalizedLeafModel,
    "stable_zero_after": StableLeafModel,
    "base_normal": BaseNormalizedLeafModel,
    "base_meaning": BaseMeaningLeafModel,
    "base_mix": BaseMixLeafModel
}



import pandas as pd
from ..constants import TIME_NAME, CENS_NAME, get_y
from os.path import dirname, join


def str_to_categ(df_col):
    uniq = df_col.unique()
    return df_col.map(dict(zip(uniq, range(len(uniq)))))


dir_env = join(dirname(__file__), "data", "BACKBLAZE")


def load_smart_2017_date_not_all_year():
    df = pd.read_csv(join(dir_env, 'smart_2017_date_not_all_year.csv'))
    df = df.rename({"time": TIME_NAME, "event": CENS_NAME}, axis=1)
    df = df.drop(columns=['serial_number', 'failure', 'date'])
    categ = ['model']
    df['model'] = str_to_categ(df['model'])
    sign = sorted(list(set(df.columns) - {CENS_NAME, TIME_NAME}))
    y = get_y(cens=df[CENS_NAME], time=df[TIME_NAME])
    X = df.loc[:, sign].reset_index()
    return X, y, sign, categ, []


def load_backblaze_dataset():  # load_smart_2017_date_not_in_last_report():
    df = pd.read_csv(join(dir_env, 'smart_2017_date_not_in_last_report.csv'))
    df = df.rename({"time": TIME_NAME, "event": CENS_NAME}, axis=1)
    df = df.drop(columns=['serial_number', 'failure', 'date'])
    categ = ['model']
    df['model'] = str_to_categ(df['model'])
    sign = sorted(list(set(df.columns) - {CENS_NAME, TIME_NAME}))
    y = get_y(cens=df[CENS_NAME], time=df[TIME_NAME])
    X = df.loc[:, sign].reset_index()
    return X, y, sign, categ, []


def load_smart_2017_raw_9_not_all_year():
    df = pd.read_csv(join(dir_env, 'smart_2017_raw_9_not_all_year.csv'))
    df = df.rename({"time": TIME_NAME, "event": CENS_NAME}, axis=1)
    df = df.drop(columns=['serial_number', 'failure', 'date'])
    categ = ['model']
    df['model'] = str_to_categ(df['model'])
    sign = sorted(list(set(df.columns) - {CENS_NAME, TIME_NAME}))
    y = get_y(cens=df[CENS_NAME], time=df[TIME_NAME])
    X = df.loc[:, sign].reset_index()
    return X, y, sign, categ, []


def load_smart_2017_raw_9_not_in_last_report():
    df = pd.read_csv(join(dir_env, 'smart_2017_raw_9_not_in_last_report.csv'))
    df = df.rename({"time": TIME_NAME, "event": CENS_NAME}, axis=1)
    df = df.drop(columns=['serial_number', 'failure', 'date'])
    categ = ['model']
    df['model'] = str_to_categ(df['model'])
    sign = sorted(list(set(df.columns) - {CENS_NAME, TIME_NAME}))
    y = get_y(cens=df[CENS_NAME], time=df[TIME_NAME])
    X = df.loc[:, sign].reset_index()
    return X, y, sign, categ, []



import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from joblib import Parallel, delayed
from ..criteria import chi2_isf

from .find_split_hist import hist_best_attr_split
from .. import constants as cnt
from ..external import LEAF_MODEL_DICT, LeafModel
from ..scheme import Scheme

# custom_params = {"axes.spines.right": False, 'grid.color': 'lightgray', 'axes.grid': True, "axes.spines.top": False}
custom_params = {"font.size": 20, "axes.labelsize": 20, "xtick.labelsize": 20, "ytick.labelsize": 20,
                 "axes.spines.right": False, 'grid.color': 'lightgray', 'axes.grid': True, "axes.spines.top": False}
sns.set_theme(style="ticks", rc=custom_params)

""" Auxiliary functions """


class Rule(object):
    """
    Node of decision tree.
    Allow to separate data into 2 child nodes

    Attributes
    ----------
    feature : str
        Name of feature for splitting
    condition : str
        Operation for splitting
    has_nan : bool
        Flag of the missing values in node

    Methods
    -------
    get_feature : Return feature
    get_condition : Return condition
    translate: Replace rule by dictionary
    to_str : Transforming to linear form
    print : Print all attributes and descriptions

    """
    def __init__(self, feature: str, condition: str, has_nan: int):
        self.feature = feature
        self.condition = condition
        self.has_nan_ = has_nan

    def get_feature(self):
        return self.feature

    def get_condition(self):
        return self.condition

    def has_nan(self):
        return self.has_nan_

    def translate(self, describe: dict):
        """
        Rename feature in rule
        """
        self.feature = describe.get(self.feature, self.feature)

    def to_str(self):
        s = f"({self.feature}{self.condition})"
        if self.has_nan_:
            s = f"({s}| nan)"
        return s

    def print(self):
        """
        Print all attributes and descriptions
        """
        print(f"has_nan: {self.has_nan()}")
        print(f"feature: {self.get_feature()}")
        print(f"condition: {self.get_condition()}")


class Node(object):
    """
    Node of decision tree.
    Allow to separate data into 2 child nodes

    Attributes
    ----------
    df : Pandas DataFrame
        Data of the Node
    numb : int
        Number or name of the Node
    full_rule : list
        Rules from the root
    depth : int
        Distance from the root
    edges : array-like
        Numbers of child nodes
    rule_edges : array-like
        Rules for child nodes
    features : list
        Available features
    categ : list
        Names of categorical features
    woe : boolean
        Mode of categorical preparation
    is_leaf : boolean
        True if node is terminal (there are no child nodes)
    verbose : int
        Print the best split of the node
    info : dict
        Parameters for finding the best splitting
    leaf_model : LeafModel
        Stratified model by parameter <<leaf_model>> (map with LEAF_MODEL_DICT)

    Methods
    -------
    check_params : Fill empty parameters and map max_features to int
    find_best_split : Choose best split of node according to parameters
    split : Find best split of df sample and create child nodes
    set_edges: Set number of child nodes from hash table of main tree
    set_leaf : Delete child nodes and set node as terminal

    predict : Return statistic values of a data
    predict_scheme : Return all possible outcomes for additional features determination

    prepare_df_for_attr : set input values to numpy format (and fill missing features)
    get_edges : defines appropriate child nodes according to input values
    get_full_rule : convert full_rules to a string format

    get_figure : Create picture of a data (hist, survival function)
    get_description : Return common values of a data (size, depth, death, cens)

    set_dot_node : add self node to the graphviz dot
    set_dot_edges : add child nodes to the graphviz dot
    translate : Replace rules and features by dictionary

    """
    def __init__(self, df,  numb=0, full_rule=[],
                 depth=0, features=[], categ=[], woe=False,
                 verbose=0, **info):
        self.df = df
        self.numb = numb
        self.size = None
        self.full_rule = full_rule
        self.depth = depth
        self.edges = np.array([], dtype=int)
        self.rule_edges = np.array([], dtype=Rule)
        self.features = features
        self.categ = categ
        self.woe = woe
        self.is_leaf = True
        self.verbose = verbose
        self.info = info
        self.check_params()

    def check_params(self):
        self.info.setdefault("bonf", True)
        self.info.setdefault("max_features", 1.0)
        self.info.setdefault("signif", 1.1)
        self.info.setdefault("signif_stat", chi2_isf(min(self.info["signif"], 1.0), df=1))
        self.info.setdefault("thres_cont_bin_max", 100)
        self.info.setdefault("normalize", True)

        if self.info["max_features"] == "sqrt":
            self.info["max_features"] = int(np.trunc(np.sqrt(len(self.features))+0.5))
        elif isinstance(self.info["max_features"], float):
            self.info["max_features"] = int(self.info["max_features"] * len(self.features))

        self.info.setdefault("weights_feature", None)
        if not (self.info["weights_feature"] is None):
            self.info["weights"] = self.df[self.info["weights_feature"]].to_numpy()

        leaf_kwargs = {k[5:]: v for k, v in self.info.items() if (k.find("leaf_") != -1) and (k != "leaf_model")}
        leaf_kwargs.setdefault("categ", self.categ)
        self.info.setdefault("leaf_model", "base_zero_after")  # base
        if isinstance(self.info["leaf_model"], str):
            self.leaf_model = LEAF_MODEL_DICT.get(self.info["leaf_model"], "base_zero_after")(**leaf_kwargs)  # base
        elif callable(self.info["leaf_model"]):  # Check is class
            self.leaf_model = self.info["leaf_model"](**leaf_kwargs)
            if not (isinstance(self.leaf_model, LeafModel)):
                self.leaf_model = None
        else:
            self.leaf_model = None
        self.size = self.df.shape[0]

        self.leaf_model.fit(self.df)
        # self.ch = np.array(
        #     [np.mean(self.df["time"]), np.std(self.df["time"]), np.sum(self.df["cens"]) / self.df["cens"].shape[0]])

    """ GROUP FUNCTIONS: CREATE LEAVES """
    def get_comb_fast(self, features):
        """
        Create set of all triplets with two target variables and one splitting feature
        """
        d_df = dict(zip(self.df.columns, self.df.values.T))

        def create_params_f(name):
            d = self.info.copy()
            d["arr"] = np.vstack((d_df[name], d_df[cnt.CENS_NAME], d_df[cnt.TIME_NAME]))
            d["type_attr"] = ("woe" if self.woe else "categ") if name in self.categ else "cont"
            return d

        return list(map(create_params_f, features))

    def find_best_split(self):
        """
        Sort through all combinations of splitting the sample by features.
        Find a split with the highest statistical value.
        """
        numb_feats = self.info["max_features"]
        numb_feats = np.clip(numb_feats, 1, len(self.features))
        n_jobs = self.info.get("n_jobs", 1 if numb_feats < 20 else 5)

        selected_feats = list(np.random.choice(self.features, size=numb_feats, replace=False))

        args = self.get_comb_fast(selected_feats)
        # ml = np.vectorize(lambda x: hist_best_attr_split(**x))(args)
        with Parallel(n_jobs=n_jobs, verbose=self.verbose) as parallel:
            ml = parallel(delayed(hist_best_attr_split)(**a) for a in args)

        attrs = {f: ml[ind] for ind, f in enumerate(selected_feats)}

        # attr = min(attrs, key=lambda x: attrs[x]["p_value"])  # best by p-value
        attr = max(attrs, key=lambda x: attrs[x]["stat_val"])  # best by stat-val
        # attrs_gr = dict(filter(lambda x: x[1]["sign_split"] > 0, attrs.items()))
        # if len(attrs_gr) == 0:
        #     attr = min(attrs, key=lambda x: attrs[x]["p_value"])
        # else:
        #     attr = min(attrs_gr, key=lambda x: attrs_gr[x]["p_value"])
        #     if self.info["bonf"]:
        #         attrs[attr]["p_value"] = attrs[attr]["p_value"] / attrs[attr]["sign_split"]

        # if attrs[attr]["sign_split"] > 0 and self.info["bonf"]:  # suffix for simple p-value
        #     attrs[attr]["p_value"] = attrs[attr]["p_value"] / attrs[attr]["sign_split"]
        return attr, attrs[attr]

    def ind_for_nodes(self, X_attr, best_split, is_categ):
        """ Map the number of the according child node by rule and sample features """
        rule_id = best_split["pos_nan"].index(0)
        query = best_split["values"][rule_id]
        if is_categ:
            values = np.isin(X_attr, eval(query[query.find("["):]))
        else:
            values = eval("X_attr" + query)
        return np.where(values, rule_id, 1 - rule_id)

    def split(self):
        """ Find best split of df sample and create child nodes """
        node_edges = np.array([], dtype=int)
        self.rule_edges = np.array([], dtype=Rule)

        attr, best_split = self.find_best_split()
        # The best split is not significant
        if best_split["sign_split"] == 0:
            if self.verbose > 0:
                print(f'The node is terminal, best p-value: {best_split["stat_val"]}')
            return node_edges
        if self.verbose > 0:
            print('='*self.depth, best_split["stat_val"], attr)

        branch_ind = self.ind_for_nodes(self.df[attr], best_split, attr in self.categ)

        for n_b in np.unique(branch_ind):
            rule = Rule(feature=attr,
                        condition=best_split["values"][n_b],
                        has_nan=best_split["pos_nan"][n_b])
            d_node = self.df[branch_ind == n_b].copy()
            N = Node(df=d_node, full_rule=self.full_rule + [rule],
                     features=self.features, categ=self.categ,
                     depth=self.depth + 1, verbose=self.verbose, **self.info)
            node_edges = np.append(node_edges, N)
            self.rule_edges = np.append(self.rule_edges, rule)

        if self.rule_edges.shape[0] == 1:
            print(branch_ind, self.df[attr], best_split, attr in self.categ)
            raise ValueError('ERROR: Only one branch created!')

        self.df = None
        return node_edges

    def set_edges(self, edges):
        self.edges = edges
        self.is_leaf = False

    def set_leaf(self):
        self.edges = np.array([], dtype=int)
        self.is_leaf = True
        self.df = None

    def prepare_df_for_attr(self, X):
        attr = self.rule_edges[0].get_feature()
        if attr not in X.columns:
            X.loc[:, attr] = np.nan
        return X[attr].to_numpy()

    def get_edges(self, X):
        X_np = self.prepare_df_for_attr(X)
        rule_id = 1 if self.rule_edges[0].has_nan() else 0

        query = self.rule_edges[rule_id].get_condition()
        if self.rule_edges[0].get_feature() in self.categ:
            values = np.isin(X_np, eval(query[query.find("["):]))
        else:
            values = eval("X_np" + query)
        return np.where(values, self.edges[rule_id], self.edges[1-rule_id])

    def get_full_rule(self):
        return " & ".join([s.to_str() for s in self.full_rule])

    def predict(self, X, target, bins=None):
        """
        Predict target values for X data

        Parameters
        ----------
        X : Pandas dataframe
            Contain input features of events
        target : str or function
            Column name, mode or aggregate function of a leaf sample
            Column name : must be in dataset.columns
            - Return mean of feature
            Mode :
            - "surv" return survival function
            - "hazard" return cumulative hazard function
            - attribute_name return attribute value (e.g. depth, numb)
            - feature_name return aggregate statistic value for node
        bins : array-like
            Points of timeline

        Returns
        -------
        res : array-like
            Values by target
        """
        if target == "surv":
            return self.leaf_model.predict_survival_at_times(X, bins)
        elif target == "hazard":
            return self.leaf_model.predict_hazard_at_times(X, bins)
        # elif target == "ch":
        #     return np.repeat(self.ch[np.newaxis, :], X.shape[0], axis=0)
        elif target in self.__dict__:
            return np.repeat(getattr(self, target, np.nan), X.shape[0], axis=0)
        return self.leaf_model.predict_feature(X, target)

    def predict_scheme(self, X, scheme_feats):
        """
        Predict target values for X data

        Parameters
        ----------
        X : Pandas dataframe
            Contain input features of events
        scheme_feats : list
            Needed features from node

        Returns
        -------
        res : Scheme
            Aggregation entity of node information
        """
        feat_means = np.array([self.leaf_model.features_predict.get(s_f, np.nan)
                               for s_f in scheme_feats])
        times = self.leaf_model.predict_list_feature(cnt.TIME_NAME)
        cens = self.leaf_model.predict_list_feature(cnt.CENS_NAME)
        return Scheme(self.get_full_rule(), times, cens, feat_means)

    """ GROUP FUNCTIONS: VISUALIZATION """

    def get_figure(self, mode="hist", bins=None, target=cnt.CENS_NAME, save_path="", lang="en"):
        """
        Save background image for graphviz node

        Parameters
        ----------
        target : str
            Described feature
        mode : string
            - "hist" plot histogram of target
            - "kde" plot smooth density of target
            - "surv" plot survival function (doesn't use target)
        bins : list
            Points for survival function
        save_path : str
            External path to save image
        """
        plt.ioff()
        # fig, ax = plt.subplots(figsize=(8, 5))
        fig, ax = plt.subplots(figsize=(8, 6))
        if mode == "hist":
            lst = self.leaf_model.predict_list_feature(target)
            plt.hist(lst, bins=25)
            ax.set_xlim([0, np.max(lst)])
            ax.set_xlabel(f'{target}', fontsize=25)
        elif mode == "kde":
            lst = self.leaf_model.predict_list_feature(target)
            # lst = self.leaf_model.old_durs
            sns.kdeplot(lst, ax=ax, linewidth=3, fill=False)
            ax.set_xlabel(f'{cnt.TRANSLATOR.get(target, {lang: target})[lang]}', fontsize=20)
            # ax.set_ylabel(f'{cnt.TRANSLATOR["density"][lang]}', fontsize=20)
        elif mode == "surv":
            sf = self.leaf_model.predict_survival_at_times(X=None, bins=bins)
            if len(sf.shape) > 1:
                sf = sf.flat
            plt.step(bins, sf, linewidth=3)
            ax.set_xlabel(cnt.TRANSLATOR["time"][lang], fontsize=25)
            ax.set_ylabel(cnt.TRANSLATOR["sf"][lang], fontsize=25)
        plt.savefig(save_path)
        plt.close(fig)

    def get_description(self, lang="en"):
        """
        Build description for graphviz node

        Returns
        -------
        label : str
            multiline string of description
        """
        m_cens = self.leaf_model.predict_feature(X=None, feature_name=cnt.CENS_NAME)
        m_time = self.leaf_model.predict_feature(X=None, feature_name=cnt.TIME_NAME)
        if isinstance(m_cens, np.ndarray):
            m_cens = m_cens[0]
        if isinstance(m_time, np.ndarray):
            m_time = m_time[0]
        label = "\n".join([f"{cnt.TRANSLATOR['size'][lang]} = {self.leaf_model.get_shape()[0]}",
                           f"{cnt.TRANSLATOR['events'][lang]} (%) = {round(m_cens, 2)}",
                           f"{cnt.TRANSLATOR['mean time'][lang]} = {round(m_time, 2)}"])
        return label

    def set_dot_node(self, dot, path_dir="", depth=None, lang="en", **args):
        """ Set node for graphviz dot with image and label """
        if not (depth is None) and depth < self.depth:
            return dot
        img_path = path_dir + str(self.numb) + '.png'
        self.get_figure(save_path=img_path, lang=lang, **args)
        dot.node(str(self.numb), label=self.get_description(lang=lang),
                 image=img_path, fontsize='30')  # fontsize='16'
        return dot

    def set_dot_edges(self, dot):
        """ Set edges for graphviz by node structure """
        if not self.is_leaf:
            for e in range(len(self.rule_edges)):
                s = self.rule_edges[e].to_str()
                dot.edge(str(self.numb), str(self.edges[e]), label=s, fontsize='30')
        return dot

    def translate(self, describe):
        """ Rename features in node by dictionary """
        self.features = [describe.get(f, f) for f in self.features]
        self.categ = [describe.get(c, c) for c in self.categ]
        for e in range(len(self.rule_edges)):
            self.rule_edges[e].translate(describe)



import numpy as np

from ..tree import CRAID
from .. import constants as cnt
from .base_ensemble import FastBaseEnsemble
from joblib import Parallel, delayed


class BootstrapCRAID(FastBaseEnsemble):
    """
    Bootstrap aggregation (Bagging) ensemble of survival decision tree.
    On each iteration probabilities of observations change by scheme.

    Attributes
    ----------
    kwargs : dict
        Parameters for building base ensemble (look at BaseEnsemble)

    Methods
    -------
    fit : build ensemble with X, y data
    """

    def __init__(self, **kwargs):
        self.name = "BootstrapCRAID"
        super().__init__(**kwargs)

    def fit(self, X, y):
        self.features = X.columns
        X = X.reset_index(drop=True)
        X[cnt.CENS_NAME] = y[cnt.CENS_NAME].astype(np.int32)
        X[cnt.TIME_NAME] = y[cnt.TIME_NAME].astype(np.float32)

        self.X_train = X
        self.y_train = y
        self.update_params()

        for i in range(self.n_estimators):
            x_sub = self.X_train.sample(n=self.size_sample, replace=self.bootstrap, random_state=i)
            x_oob = self.X_train.loc[self.X_train.index.difference(x_sub.index), :]

            x_sub = x_sub.reset_index(drop=True)
            X_sub_tr, y_sub_tr = cnt.pd_to_xy(x_sub)

            model = CRAID(features=self.features, random_state=i, **self.tree_kwargs)
            model.fit(X_sub_tr, y_sub_tr)

            self.add_model(model, x_oob)
        print(f"fitted: {len(self.models)} models.")


class ParallelBootstrapCRAID(BootstrapCRAID):
    """
    Parallel realization of BootstrapCRAID.
    Trees are built in parallel, and the number of processes is specified via n_jobs.
    """

    @staticmethod
    def fit_tree(x_sub, params):
        """
        An atomic out-of-class tree construction operation for parallelization.
        """
        X_tr, y_tr = cnt.pd_to_xy(x_sub)
        model = CRAID(**params)
        model.fit(X_tr, y_tr)
        return model

    def fit(self, X, y):
        self.features = X.columns
        X = X.reset_index(drop=True)
        X[cnt.CENS_NAME] = y[cnt.CENS_NAME].astype(np.int32)
        X[cnt.TIME_NAME] = y[cnt.TIME_NAME].astype(np.float32)

        self.X_train = X
        self.y_train = y
        self.update_params()

        p_s = []
        x_sub_ind = []
        for i in range(self.n_estimators):
            x_sub = self.X_train.sample(n=self.size_sample, replace=self.bootstrap, random_state=i)
            x_sub_ind.append(x_sub.index)

            x_sub = x_sub.reset_index(drop=True)
            params = self.tree_kwargs.copy()
            params['random_state'] = i
            params['features'] = self.features
            params['n_jobs'] = 1
            p_s.append({"x_sub": x_sub, "params": params})

        with Parallel(n_jobs=self.tree_kwargs.get("n_jobs", 10), verbose=False) as parallel:
            ml = parallel(delayed(self.fit_tree)(**p) for p in p_s)

        for model, x_sub_ind in zip(ml, x_sub_ind):
            x_oob = self.X_train.loc[self.X_train.index.difference(x_sub_ind), :]
            self.add_model(model, x_oob)
        # print(f"fitted: {len(self.models)} models.")

    # def predict_at_times(self, x_test, bins, aggreg=True, mode="surv"):
    #     with Parallel(n_jobs=self.tree_kwargs.get("n_jobs", 10), verbose=False, batch_size=10) as parallel:
    #         ml = parallel(delayed(lambda m: m.predict_at_times(x_test, bins=bins,
    #                                                            mode=mode)[np.newaxis, :])(m) for m in self.models)
    #     res = np.vstack(ml)
    #     if aggreg:
    #         res = self.get_aggreg(res)
    #         if mode == "surv":
    #             res[:, -1] = 0
    #             res[:, 0] = 1
    #     return res
    #
    # def predict(self, x_test, aggreg=True, **kwargs):
    #     with Parallel(n_jobs=self.tree_kwargs.get("n_jobs", 10), verbose=False, batch_size=10) as parallel:
    #         ml = parallel(delayed(lambda m: m.predict(x_test, **kwargs))(m) for m in self.models)
    #     res = np.array(ml)
    #     if aggreg:
    #         res = self.get_aggreg(res)
    #     return res



import numpy as np
from .. import constants as cnt


class LeafModel(object):
    """
    Unified wrapper for external models.
    Can be used as leaf model in tree-based methods.
    Supported in Experiments module.

    Attributes
    ----------
    features : list
        Covariates for fitting
    features_predict : dict
        Description values of features (mean by default)
    lists: dict
        Source values of target variables
    weights_name : str
        Name of the weighting column
    kwargs : dict
        Internal parameters for the leaf model

    Methods
    -------
    fit : build ensemble on source X_node data
    predict_list_feature : return full data of variables
    predict_feature : return aggregated data of variables
    predict_survival_at_times : return survival function by bins
    predict_hazard_at_times : return hazard function by bins

    """
    def __init__(self, features=[], categ=[], weights_name=None, **kwargs):
        self.kwargs = kwargs
        self.features = features
        self.categ = categ
        self.shape = None
        self.features_predict = dict()
        self.lists = dict()
        self.weights_name = weights_name
        self.weights = None

    def fit(self, X_node, *args, **kwargs):
        if not self.features:
            self.features = X_node.columns
        # self.features = sorted(list(set(self.features + [cnt.TIME_NAME, cnt.CENS_NAME])))
        # X_sub = X_node[self.features]
        # self.shape = X_sub.shape
        # self.features_predict = X_sub.mean(axis=0).to_dict()
        # self.lists = X_sub[[cnt.TIME_NAME, cnt.CENS_NAME]].to_dict(orient="list")

        self.shape = (X_node.shape[0], len(self.features))
        l = dict(zip(X_node.columns, X_node.values.T))
        self.lists = {k: v for k, v in l.items() if k in [cnt.TIME_NAME, cnt.CENS_NAME]}
        self.features_predict = {k: cnt.mode(v) if k in self.categ else np.mean(v) for k, v in l.items() if k in self.features}

        if self.weights_name is None:
            self.weights = None
        else:
            if self.weights_name in X_node.columns:
                self.weights = X_node[self.weights_name].to_numpy()
            else:
                self.weights = np.ones(self.shape[0])

    def get_shape(self):
        return self.shape

    def predict_list_feature(self, feature_name):
        if feature_name in self.lists.keys():
            return np.array(self.lists[feature_name])
        return None

    def predict_feature(self, X=None, feature_name=None):
        value = self.features_predict.get(feature_name)
        if X is None:
            return value
        return np.repeat(value, X.shape[0], axis=0)

    def predict_survival_at_times(self, X=None, bins=None):
        pass

    def predict_hazard_at_times(self, X=None, bins=None):
        pass


class NonparamLeafModel(LeafModel):
    survival_class = None
    hazard_class = None

    def __init__(self, *args, **kwargs):
        if (self.survival_class is None) and (self.hazard_class is None):
            raise Exception("There is no survival or hazard base class!")
        self.survival = None
        self.hazard = None
        super().__init__(*args, **kwargs)

    def predict_survival_at_times(self, X=None, bins=None):
        if self.survival_class is None:
            hf = self.predict_hazard_at_times(X=X, bins=bins)
            sf = np.exp(-1 * hf)
            return sf

        if self.survival is None:
            self.survival = self.survival_class()
            self.survival.fit(self.lists[cnt.TIME_NAME],
                              self.lists[cnt.CENS_NAME],
                              self.weights)
        if bins is None:
            bins = self.survival.timeline
        sf = self.survival.survival_function_at_times(bins)
        if type(sf).__name__ == "Series":
            sf = sf.values
        if X is None:
            return sf
        return np.repeat(sf[np.newaxis, :], X.shape[0], axis=0)

    def predict_hazard_at_times(self, X=None, bins=None):
        if self.hazard_class is None:
            sf = self.predict_survival_at_times(X=X, bins=bins)
            hf = -1 * np.log(sf + 1e-100)
            return hf

        if self.hazard is None:
            self.hazard = self.hazard_class()
            self.hazard.fit(self.lists[cnt.TIME_NAME],
                            self.lists[cnt.CENS_NAME],
                            self.weights)
        if bins is None:
            bins = self.hazard.timeline
        hf = self.hazard.cumulative_hazard_at_times(bins)
        if type(hf).__name__ == "Series":
            hf = hf.values
        if X is None:
            return hf
        return np.repeat(hf[np.newaxis, :], X.shape[0], axis=0)


class NormalizedLeafModel(NonparamLeafModel):
    """
    Nonparametric model with virtual events (normal distribution) and non-informative censuring
    """
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)

        durs = np.random.normal(np.mean(self.lists[cnt.TIME_NAME]),
                                np.std(self.lists[cnt.TIME_NAME]) / np.sqrt(2), 1000)
        events = np.random.choice(self.lists[cnt.CENS_NAME], size=1000, replace=True)
        self.lists[cnt.CENS_NAME] = events[durs > 0]
        self.lists[cnt.TIME_NAME] = durs[durs > 0]
        self.weights = np.ones_like(self.lists[cnt.TIME_NAME])


class StableLeafModel(NonparamLeafModel):
    """
    Nonparametric model with stable virtual events (with median mu and adaptive size) and non-informative censuring
    """
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)

        durs = np.random.normal(np.median(self.lists[cnt.TIME_NAME]),
                                np.std(self.lists[cnt.TIME_NAME]) / np.sqrt(2), self.shape[0]*5)
        events = np.random.choice(self.lists[cnt.CENS_NAME], size=self.shape[0]*5, replace=True)
        durs = np.hstack([self.lists[cnt.TIME_NAME], durs])
        events = np.hstack([self.lists[cnt.CENS_NAME], events])

        self.lists[cnt.CENS_NAME] = events[durs > 0]
        self.lists[cnt.TIME_NAME] = durs[durs > 0]
        self.weights = np.ones_like(self.lists[cnt.TIME_NAME])


class MeaningLeafModel(NonparamLeafModel):
    """
    Nonparametric model with virtual events generated by bootstrap meaning
    """
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)
        self.lists[cnt.TIME_NAME] = np.random.choice(self.lists[cnt.TIME_NAME], size=(2, 1000), replace=True).mean(axis=0)
        self.lists[cnt.CENS_NAME] = np.random.choice(self.lists[cnt.CENS_NAME], size=1000, replace=True)
        self.weights = np.ones_like(self.lists[cnt.TIME_NAME])


class MixLeafModel(NormalizedLeafModel):
    """
    Nonparametric model with virtual events (normal distributions + bootstrap meaning)
    """
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)
        self.lists[cnt.TIME_NAME] = np.random.choice(self.lists[cnt.TIME_NAME], size=(2, 1000), replace=True).mean(axis=0)
        self.lists[cnt.CENS_NAME] = np.random.choice(self.lists[cnt.CENS_NAME], size=1000, replace=True)
        self.weights = np.ones_like(self.lists[cnt.TIME_NAME])



import numpy as np

from .. import constants as cnt
from .boosting import BoostingCRAID


def loglikelihood_i(time, cens, sf, cumhf, bins):
    index_times = np.digitize(time, bins, right=True) - 1
    hf = np.hstack((cumhf[:, 0][np.newaxis].T, np.diff(cumhf)))
    sf_by_times = np.take_along_axis(sf, index_times[:, np.newaxis], axis=1)[:, 0] + 1e-10
    hf_by_times = (np.take_along_axis(hf, index_times[:, np.newaxis], axis=1)[:, 0] + 1e-10) ** cens
    return np.log(sf_by_times) + np.log(hf_by_times)


def values_to_hist(values):
    unq, idx = np.unique(values, return_inverse=True)
    # calculate the weighted frequencies of these indices
    freqs_idx = np.bincount(idx)
    # reconstruct the array of frequencies of the elements
    return freqs_idx[idx]


class ProbBoostingCRAID(BoostingCRAID):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "ProbBoostingCRAID"

    def count_model_weights(self, model, X_sub, y_sub):
        if self.all_weight:
            X_sub = self.X_train
            y_sub = self.y_train

        pred_sf = model.predict_at_times(X_sub, bins=self.bins, mode="surv")
        pred_hf = model.predict_at_times(X_sub, bins=self.bins, mode="hazard")

        time_hist = values_to_hist(y_sub[cnt.TIME_NAME])
        lp_ti = np.log(time_hist / y_sub[cnt.TIME_NAME].shape)
        lp_xi_ti = np.log(1 / time_hist)
        likel = loglikelihood_i(y_sub[cnt.TIME_NAME], y_sub[cnt.CENS_NAME], pred_sf, pred_hf, self.bins)

        lp_xi = lp_ti + lp_xi_ti - likel
        wei = - np.exp(-lp_xi)
        betta = np.sum(likel)
        return wei, betta

    def update_weight(self, index, wei_i):
        if self.all_weight:
            self.weights = self.weights + wei_i
        else:
            self.weights[index] = (self.weights[index] + wei_i)
        self.weights = (self.weights - self.weights.min())
        # self.weights = (self.weights - self.weights.min()) / (self.weights.max() - self.weights.min())

    def get_aggreg(self, x):
        if self.aggreg_func == 'median':
            return np.median(x, axis=0)
        elif self.aggreg_func == 'wei':
            inv_wei = -1/np.array(self.bettas)
            wei = inv_wei/sum(inv_wei)
            return np.sum((x.T*wei).T, axis=0)
        return np.mean(x, axis=0)



import numpy as np

from .. import metrics as metr
from .. import constants as cnt

from ..tree import CRAID
from .boosting import BoostingCRAID
import matplotlib.pyplot as plt


class IBSCleverBoostingCRAID(BoostingCRAID):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IBSCleverBoostingCRAID"

    def fit(self, X, y):
        self.features = X.columns
        X = X.reset_index(drop=True)
        X[cnt.CENS_NAME] = y[cnt.CENS_NAME].astype(np.int32)
        X[cnt.TIME_NAME] = y[cnt.TIME_NAME].astype(np.float32)

        self.X_train = X
        self.X_train["ind_start"] = self.X_train.index
        self.y_train = y

        self.weights = np.ones(self.X_train.shape[0], dtype=float)
        self.bettas = []
        self.l_ibs = []
        self.l_weights = []
        self.update_params()

        for i in range(self.n_estimators):
            x_sub = self.X_train.sample(n=self.size_sample,
                                        # weights=self.weights,
                                        replace=self.bootstrap, random_state=i)

            x_oob = self.X_train.loc[self.X_train.index.difference(x_sub.index), :]
            # print(f"UNIQUE ({i}):{np.unique(x_sub.index).shape[0]}, DIST:", np.bincount(x_sub["cens"]))
            x_sub = x_sub.reset_index(drop=True)
            X_sub_tr, y_sub_tr = cnt.pd_to_xy(x_sub)
            if self.weighted_tree:
                X_sub_tr["weights_obs"] = self.weights[x_sub['ind_start']]

            #             plt.scatter(y_sub_tr["time"], self.weights[x_sub['ind_start']])
            #             plt.show()

            model = CRAID(features=self.features, random_state=i, **self.tree_kwargs)
            model.fit(X_sub_tr, y_sub_tr)

            wei_i, betta_i = self.count_model_weights(model, X_sub_tr, y_sub_tr)
            self.add_model(model, x_oob, wei_i, betta_i)
            self.update_weight(x_sub['ind_start'], wei_i)

    def predict(self, x_test, aggreg=True, **kwargs):
        res = []
        weights = []
        for i in range(len(self.models)):
            res.append(self.models[i].predict(x_test, **kwargs))

        res = np.array(res)
        weights = None  # np.vstack(weights).T
        if aggreg:
            res = self.get_aggreg(res, weights)
        return res

    def predict_at_times(self, x_test, bins, aggreg=True, mode="surv"):
        res = []
        weights = []
        for i in range(len(self.models)):
            res.append(self.models[i].predict_at_times(x_test, bins=bins,
                                                       mode=mode))

        res = np.array(res)
        weights = None  # np.vstack(weights).T
        if aggreg:
            res = self.get_aggreg(res, weights)
            if mode == "surv":
                res[:, -1] = 0
                res[:, 0] = 1
        return res

    def count_model_weights(self, model, X_sub, y_sub):
        if self.all_weight:
            X_sub = self.X_train
            y_sub = self.y_train
        pred_sf = model.predict_at_times(X_sub, bins=self.bins, mode="surv")

        # PRED WEI!!!
        ibs_sf = metr.ibs_remain(self.y_train, y_sub, pred_sf, self.bins, axis=0)

        #         if len(self.bettas) > 0:
        #             pred_ens = self.predict_at_times(X_sub, bins=self.bins, mode="surv")
        #             ibs_ens = metr.ibs_WW(self.y_train, y_sub, pred_ens, self.bins)
        #             betta = ibs_ens / np.mean(ibs_sf)
        #         else:
        #             betta = 1
        #         wei = ibs_sf

        if len(self.bettas) > 0:
            pred_ens = self.predict_at_times(X_sub, bins=self.bins, mode="surv")

            ibs_sf_all = metr.ibs_remain(self.y_train, y_sub, pred_sf, self.bins)
            ibs_ens = metr.ibs_remain(self.y_train, y_sub, pred_ens, self.bins)

            betta = ibs_ens / ibs_sf_all
            # betta = ibs_ens / np.mean(ibs_sf)
            # print(betta)
            wei = (ibs_ens + (betta ** 2) * ibs_sf) / (1 + betta) ** 2
        else:
            betta = 1
            wei = ibs_sf
        return wei, abs(betta)

    def update_weight(self, index, wei_i):
        # PRED WEI!!!
        #         if len(self.models) > 1:
        #             self.weights = self.weights + (self.bettas[-1]**2) * wei_i
        #             self.weights /= (1 + self.bettas[-1])**2
        #             self.bettas = list(np.array(self.bettas)/np.sum(self.bettas))
        #         else:
        #             self.weights = wei_i
        self.weights = wei_i

    def get_aggreg(self, x, wei=None):
        if self.aggreg_func == 'median':
            return np.median(x, axis=0)
        elif self.aggreg_func == "wei":
            if wei is None:
                wei = np.array(self.bettas)
            wei = wei / np.sum(wei)
            return np.sum((x.T * wei).T, axis=0)
        elif self.aggreg_func == "argmean":
            wei = np.where(np.argsort(np.argsort(wei, axis=1), axis=1) > len(self.bettas) // 2, 1, 0)
            wei = wei / np.sum(wei, axis=1).reshape(-1, 1)
            return np.sum((x.T * wei).T, axis=0)
        elif self.aggreg_func == "argwei":
            wei = np.where(np.argsort(np.argsort(wei, axis=1), axis=1) > len(self.bettas) // 2,
                           1 / np.array(self.bettas), 0)
            wei = wei / np.sum(wei, axis=1).reshape(-1, 1)
            return np.sum((x.T * wei).T, axis=0)
        return np.mean(x, axis=0)

    def plot_curve(self, X_tmp, y_tmp, bins, label="", metric="ibs"):
        res = []
        metr_vals = []
        for i in range(len(self.models)):
            res.append(self.models[i].predict_at_times(X_tmp, bins=bins, mode="surv"))

            res_all = np.array(res)
            res_all = self.get_aggreg(res_all, np.array(self.bettas)[:i + 1])
            res_all[:, -1] = 0
            res_all[:, 0] = 1
            if metric == "ibs":
                metr_vals.append(metr.ibs_WW(self.y_train, y_tmp, res_all, bins))
            else:
                metr_vals.append(metr.auprc(self.y_train, y_tmp, res_all, bins))
        plt.plot(range(len(self.models)), metr_vals, label=label)
